from DungeonGenerator.player_lib import *
from Fortuna import MultiChoice


get_archetype = MultiChoice(
    "\nChoose an archetype:",
    options=("Apprentice", "Caster", "Melee", "Ranged", "Healer"),
    strict=True,
)
get_caster = MultiChoice(
    "\nWhat type of caster?",
    options=("Wizard", "Sorcerer", "Warlock", "Cleric", "Druid", "Bard"),
    strict=True,
)
get_melee = MultiChoice(
    "\nWhat type of melee?",
    options=("Fighter", "Barbarian", "Knight", "Rogue", "Monk", "Paladin", "Bard"),
    strict=True,
)
get_ranged = MultiChoice(
    "\nWhat type of ranged?",
    options=("Ranger", "Rogue", "Fighter", "Bard"),
    strict=True,
)
get_healer = MultiChoice(
    "\nWhat type of healer?",
    options=("Cleric", "Druid", "Paladin", "Bard"),
    strict=True,
)

class_hierarchy = {
    "Apprentice": lambda: "Apprentice",
    "Caster": get_caster,
    "Melee": get_melee,
    "Ranged": get_ranged,
    "Healer": get_healer,
}

get_race = MultiChoice(
    "\nChoose a race:",
    options=(
        "Human", "Mountain Dwarf", "Hill Dwarf", "Shout Halfling",
        "Lightfoot Halfling", "Forest Gnome", "Rock Gnome", "High Elf",
        "Wood Elf", "Half-elf", "Drow", "Half-orc", "Dragonborn", "Tiefling",
    ),
    strict=True,
)


class PickNumber:

    def __init__(self, query=None, target_range=None, default=None):
        target = target_range or range(1, 101)
        que = query or "Pick a number"
        self.query = f"{que}: {target[0]}-{target[-1]}"
        self.target_range = target
        self.default = default
        self.cursor = ">>> "
        self.choice_pack = (
            self.query,
            self.cursor,
        )

    def _get_answer(self):
        return input('\n'.join(self.choice_pack)) or self.default

    def __call__(self):
        answer = self._get_answer()
        if not answer:
            return random_value(self.target_range)
        elif int(answer) in self.target_range:
            return int(answer)
        else:
            return self()


def make_character():
    my_race = get_race()
    get_class = class_hierarchy[get_archetype()]
    my_class = get_class()
    get_level = PickNumber("\nChoose a level", range(1, 21))
    level = get_level() if my_class != 'Apprentice' else 1
    return Player(eval(my_class), race=my_race, level=level)


if __name__ == '__main__':
    print(f"\n{make_character()}")



from Fortuna import *
from DungeonGenerator.utilities import cr_adapter
from DungeonGenerator.spells import random_spell


def magic_table_by_cr(cr):
    tables = (magic_table_f, magic_table_g, magic_table_h, magic_table_i)
    tier = smart_clamp(cr // 5, 0, 3)
    return tables[tier]()


def get_loot(cr, monster_name="Any", room_name="Any", is_villain=False):
    loot = get_monster_loot(cr)
    epic_monsters = ("Dragon", "Mimic", "Gelatinous", "Legendary")
    epic_rooms = ("Lair", "Fortuna", "Dungeon Heart", "Treasury")
    if is_villain or any(m in monster_name for m in epic_monsters) or any(r in room_name for r in epic_rooms):
        loot += get_villain_loot(cr)
    return loot


def get_monster_loot(cr):
    if cr <= 4:
        loot = RankOneLoot
    elif cr <= 10:
        loot = RankTwoLoot
    elif cr <= 16:
        loot = RankThreeLoot
    else:
        loot = RankFourLoot
    return loot()


def get_villain_loot(cr):
    if cr <= 4:
        loot = RankOneHoardLoot
    elif cr <= 10:
        loot = RankTwoHoardLoot
    elif cr <= 16:
        loot = RankThreeHoardLoot
    else:
        loot = RankFourHoardLoot
    return loot()


class Loot:
    """ Base Class for loot, starts empty """

    def __init__(self):
        self.coinage = {
            'Copper': 0, 'Silver': 0, 'Electrum': 0, 'Gold': 0, 'Platinum': 0,
            'Gems': 0, 'Jewels': 0
        }
        self.magic_items = []

    def get_gold_value(self):
        golds = (
            self.coinage["Gold"],
            self.coinage["Gems"],
            self.coinage["Jewels"],
            self.coinage["Platinum"] * 10,
            self.coinage["Electrum"] / 2,
            self.coinage["Silver"] / 10,
            self.coinage["Copper"] / 100,
        )
        return sum(golds)

    def __add__(self, other):
        new_loot = Loot()
        for val in self.coinage:
            new_loot.coinage[val] = self.coinage[val] + other.coinage[val]
        new_loot.magic_items.extend(self.magic_items + other.magic_items)
        return new_loot

    def __repr__(self):
        return self.reduce_loot()

    def __len__(self):
        return len(self.magic_items)

    def reduce_loot(self):
        output = []
        for coin in self.coinage:
            if self.coinage[coin] > 0:
                if coin == "Gems" or coin == "Jewels":
                    output.append(f"{coin} {self.coinage[coin]} GP")
                else:
                    output.append(f"{self.coinage[coin]} {coin} Coins")
        if self.magic_items:
            for itm in sorted(list(set(self.magic_items))):
                output.append(itm)
        if len(output) == 0:
            output = ['Empty']
        return ', '.join(output)

    def set_loot(self, loot_data):
        loot = cumulative_weighted_choice(loot_data)
        for key in loot:
            if key == "Magic_Items":
                for itm in loot[key]:
                    table, amount = itm
                    if amount == 1:
                        self.magic_items.append(table())
                    else:
                        self.magic_items.extend(table() for _ in range(dice(*amount)))
            else:
                coin_type, amount = key, loot[key]
                n = len(amount)
                if n == 2:
                    r, s = amount
                    self.coinage[coin_type] += dice(r, s)
                elif n == 3:
                    r, s, m = amount
                    self.coinage[coin_type] += dice(r, s) * m


class RankOneLoot(Loot):
    """ Monster CR 0-4 """

    def __init__(self):
        super().__init__()
        loot_data = (
            (30, {"Copper": (5, 6)}),
            (60, {"Silver": (4, 6)}),
            (70, {"Electrum": (3, 6)}),
            (95, {"Gold": (3, 6)}),
            (100, {"Platinum": (1, 6)}),
        )
        self.set_loot(loot_data)


class RankTwoLoot(Loot):
    """ Monster CR 5-10 """

    def __init__(self):
        super().__init__()
        loot_data = (
            (30, {"Copper": (4, 6, 100), "Electrum": (1, 6, 10)}),
            (60, {"Silver": (6, 6, 100), "Gold": (2, 6, 10)}),
            (70, {"Electrum": (3, 6, 10), "Gold": (2, 6, 10)}),
            (95, {"Gold": (4, 6, 10)}),
            (100, {"Gold": (2, 6, 10), "Platinum": (3, 6)}),
        )
        self.set_loot(loot_data)


class RankThreeLoot(Loot):
    """ Monster CR 11-16 """

    def __init__(self):
        super().__init__()
        loot_data = (
            (20, {"Silver": (4, 6, 100), "Gold": (1, 6, 100)}),
            (35, {"Electrum": (1, 6, 100), "Gold": (1, 6, 100)}),
            (75, {"Gold": (2, 6, 100), "Platinum": (1, 6, 10)}),
            (100, {"Gold": (2, 6, 100), "Platinum": (2, 6, 10)}),
        )
        self.set_loot(loot_data)


class RankFourLoot(Loot):
    """ Monster CR 17+ """

    def __init__(self):
        super().__init__()
        loot_data = (
            (15, {"Electrum": (2, 6, 1000), "Gold": (8, 6, 100)}),
            (55, {"Gold": (1, 6, 1000), "Platinum": (1, 6, 100)}),
            (100, {"Gold": (1, 6, 1000), "Platinum": (2, 6, 100)})
        )
        self.set_loot(loot_data)


class RankOneHoardLoot(Loot):
    """ Hoard CR 0-4 """

    def __init__(self):
        super().__init__()
        self.coinage = {
            'Copper': dice(6, 6) * 100, 'Silver': dice(3, 6) * 100,
            'Electrum': 0, 'Gold': dice(2, 6) * 10, 'Platinum': 0,
            'Gems': 0, 'Jewels': 0
        }
        loot_data = (
            (6, {}),
            (16, {"Gems": (2, 6, 10)}),
            (26, {"Jewels": (2, 4, 25)}),
            (36, {"Gems": (2, 6, 50)}),
            (44, {"Gems": (2, 6, 10), "Magic_Items": [(magic_table_a, (1, 6))]}),
            (52, {"Jewels": (2, 4, 25), "Magic_Items": [(magic_table_a, (1, 6))]}),
            (60, {"Gems": (2, 6, 50), "Magic_Items": [(magic_table_a, (1, 6))]}),
            (65, {"Gems": (2, 6, 10), "Magic_Items": [(magic_table_b, (1, 4))]}),
            (70, {"Jewels": (2, 4, 25), "Magic_Items": [(magic_table_b, (1, 4))]}),
            (75, {"Gems": (2, 6, 50), "Magic_Items": [(magic_table_b, (1, 4))]}),
            (78, {"Gems": (2, 6, 10), "Magic_Items": [(magic_table_c, (1, 4))]}),
            (80, {"Jewels": (2, 4, 25), "Magic_Items": [(magic_table_c, (1, 4))]}),
            (85, {"Gems": (2, 6, 50), "Magic_Items": [(magic_table_c, (1, 4))]}),
            (92, {"Gems": (2, 4, 25), "Magic_Items": [(magic_table_f, (1, 4))]}),
            (97, {"Jewels": (2, 6, 50), "Magic_Items": [(magic_table_f, (1, 4))]}),
            (99, {"Gems": (2, 4, 25), "Magic_Items": [(magic_table_g, 1)]}),
            (100, {"Jewels": (2, 6, 50), "Magic_Items": [(magic_table_g, 1)]})
        )
        self.set_loot(loot_data)


class RankTwoHoardLoot(Loot):
    """ Hoard CR 5-10 """

    def __init__(self):
        super().__init__()
        self.coinage = {
            'Copper': dice(2, 6) * 100, 'Silver': dice(2, 6) * 1000,
            'Electrum': 0, 'Gold': dice(6, 6) * 100, 'Platinum': dice(3, 6) * 10,
            'Gems': 0, 'Jewels': 0
        }
        loot_data = (
            (4, {}),
            (10, {"Jewels": (2, 4, 25)}),
            (16, {"Gems": (3, 6, 50)}),
            (22, {"Gems": (3, 6, 100)}),
            (28, {"Jewels": (2, 4, 250)}),
            (32, {"Jewels": (2, 4, 25), "Magic_Items": [(magic_table_a, (1, 6))]}),
            (36, {"Gems": (3, 6, 50), "Magic_Items": [(magic_table_a, (1, 6))]}),
            (40, {"Gems": (3, 6, 100), "Magic_Items": [(magic_table_a, (1, 6))]}),
            (44, {"Jewels": (2, 4, 250), "Magic_Items": [(magic_table_a, (1, 6))]}),
            (49, {"Jewels": (2, 4, 25), "Magic_Items": [(magic_table_b, (1, 4))]}),
            (54, {"Gems": (3, 6, 50), "Magic_Items": [(magic_table_b, (1, 4))]}),
            (59, {"Gems": (3, 6, 100), "Magic_Items": [(magic_table_b, (1, 4))]}),
            (63, {"Jewels": (2, 4, 250), "Magic_Items": [(magic_table_b, (1, 4))]}),
            (66, {"Jewels": (2, 4, 25), "Magic_Items": [(magic_table_c, (1, 4))]}),
            (69, {"Gems": (3, 6, 50), "Magic_Items": [(magic_table_c, (1, 4))]}),
            (72, {"Gems": (3, 6, 100), "Magic_Items": [(magic_table_c, (1, 4))]}),
            (74, {"Jewels": (2, 4, 250), "Magic_Items": [(magic_table_c, (1, 4))]}),
            (76, {"Jewels": (2, 4, 25), "Magic_Items": [(magic_table_d, 1)]}),
            (78, {"Gems": (3, 6, 50), "Magic_Items": [(magic_table_d, 1)]}),
            (79, {"Gems": (3, 6, 100), "Magic_Items": [(magic_table_d, 1)]}),
            (80, {"Jewels": (2, 4, 250), "Magic_Items": [(magic_table_d, 1)]}),
            (84, {"Jewels": (2, 4, 25), "Magic_Items": [(magic_table_f, (1, 4))]}),
            (88, {"Gems": (3, 6, 50), "Magic_Items": [(magic_table_f, (1, 4))]}),
            (91, {"Gems": (3, 6, 100), "Magic_Items": [(magic_table_f, (1, 4))]}),
            (94, {"Jewels": (2, 4, 250), "Magic_Items": [(magic_table_f, (1, 4))]}),
            (96, {"Gems": (2, 4, 100), "Magic_Items": [(magic_table_g, (1, 4))]}),
            (98, {"Jewels": (2, 4, 250), "Magic_Items": [(magic_table_g, (1, 4))]}),
            (99, {"Gems": (3, 6, 100), "Magic_Items": [(magic_table_h, 1)]}),
            (100, {"Jewels": (2, 4, 250), "Magic_Items": [(magic_table_h, 1)]})
        )
        self.set_loot(loot_data)


class RankThreeHoardLoot(Loot):
    """ Hoard CR 11-16 """

    def __init__(self):
        super().__init__()
        self.coinage = {
            'Copper': 0, 'Silver': 0, 'Electrum': 0,
            'Gold': dice(4, 6) * 1000, 'Platinum': dice(5, 6) * 100,
            'Gems': 0, 'Jewels': 0
        }
        loot_data = (
            (3, {}),
            (6, {"Jewels": (2, 4, 250)}),
            (9, {"Jewels": (2, 4, 750)}),
            (12, {"Gems": (3, 6, 500)}),
            (15, {"Gems": (3, 6, 1000)}),
            (19, {"Jewels": (2, 4, 250), "Magic_Items": [(magic_table_a, (1, 4)), (magic_table_b, (1, 6))]}),
            (23, {"Jewels": (2, 4, 750), "Magic_Items": [(magic_table_a, (1, 4)), (magic_table_b, (1, 6))]}),
            (26, {"Gems": (3, 6, 500), "Magic_Items": [(magic_table_a, (1, 4)), (magic_table_b, (1, 6))]}),
            (29, {"Gems": (3, 6, 1000), "Magic_Items": [(magic_table_a, (1, 4)), (magic_table_b, (1, 6))]}),
            (35, {"Jewels": (2, 4, 250), "Magic_Items": [(magic_table_c, (1, 6))]}),
            (40, {"Jewels": (2, 4, 750), "Magic_Items": [(magic_table_c, (1, 6))]}),
            (45, {"Gems": (3, 6, 500), "Magic_Items": [(magic_table_c, (1, 6))]}),
            (50, {"Gems": (3, 6, 1000), "Magic_Items": [(magic_table_c, (1, 6))]}),
            (54, {"Jewels": (2, 4, 250), "Magic_Items": [(magic_table_d, (1, 4))]}),
            (58, {"Jewels": (2, 4, 750), "Magic_Items": [(magic_table_d, (1, 4))]}),
            (62, {"Gems": (3, 6, 500), "Magic_Items": [(magic_table_d, (1, 4))]}),
            (66, {"Gems": (3, 6, 1000), "Magic_Items": [(magic_table_d, (1, 4))]}),
            (68, {"Jewels": (2, 4, 250), "Magic_Items": [(magic_table_e, 1)]}),
            (70, {"Jewels": (2, 4, 750), "Magic_Items": [(magic_table_e, 1)]}),
            (72, {"Gems": (3, 6, 500), "Magic_Items": [(magic_table_e, 1)]}),
            (74, {"Gems": (3, 6, 1000), "Magic_Items": [(magic_table_e, 1)]}),
            (76, {"Jewels": (2, 4, 250), "Magic_Items": [(magic_table_g, (1, 4))]}),
            (78, {"Jewels": (2, 4, 750), "Magic_Items": [(magic_table_g, (1, 4))]}),
            (80, {"Gems": (3, 6, 500), "Magic_Items": [(magic_table_g, (1, 4))]}),
            (82, {"Gems": (3, 6, 1000), "Magic_Items": [(magic_table_g, (1, 4))]}),
            (85, {"Jewels": (2, 4, 250), "Magic_Items": [(magic_table_h, (1, 4))]}),
            (88, {"Jewels": (2, 4, 750), "Magic_Items": [(magic_table_h, (1, 4))]}),
            (90, {"Gems": (3, 6, 500), "Magic_Items": [(magic_table_h, (1, 4))]}),
            (92, {"Gems": (3, 6, 1000), "Magic_Items": [(magic_table_h, (1, 4))]}),
            (94, {"Jewels": (2, 4, 250), "Magic_Items": [(magic_table_i, 1)]}),
            (96, {"Jewels": (2, 4, 750), "Magic_Items": [(magic_table_i, 1)]}),
            (98, {"Gems": (3, 6, 500), "Magic_Items": [(magic_table_i, 1)]}),
            (100, {"Gems": (3, 6, 1000), "Magic_Items": [(magic_table_i, 1)]}),
        )
        self.set_loot(loot_data)


class RankFourHoardLoot(Loot):
    """ Hoard CR 17+ """

    def __init__(self):
        super().__init__()
        self.coinage = {
            'Copper': 0, 'Silver': 0, 'Electrum': 0,
            'Gold': dice(12, 6) * 1000, 'Platinum': dice(8, 6) * 1000,
            'Gems': 0, 'Jewels': 0
        }
        loot_data = (
            (2, {}),
            (5, {"Gems": (3, 6, 1000), "Magic_Items": [(magic_table_c, (1, 8))]}),
            (8, {"Jewels": (1, 10, 2500), "Magic_Items": [(magic_table_c, (1, 8))]}),
            (11, {"Jewels": (1, 4, 7500), "Magic_Items": [(magic_table_c, (1, 8))]}),
            (14, {"Gems": (1, 8, 5000), "Magic_Items": [(magic_table_c, (1, 8))]}),
            (22, {"Gems": (3, 6, 1000), "Magic_Items": [(magic_table_d, (1, 6))]}),
            (30, {"Jewels": (1, 10, 2500), "Magic_Items": [(magic_table_d, (1, 6))]}),
            (38, {"Jewels": (1, 4, 7500), "Magic_Items": [(magic_table_d, (1, 6))]}),
            (46, {"Gems": (1, 8, 5000), "Magic_Items": [(magic_table_d, (1, 6))]}),
            (52, {"Gems": (3, 6, 1000), "Magic_Items": [(magic_table_e, (1, 6))]}),
            (58, {"Jewels": (1, 10, 2500), "Magic_Items": [(magic_table_e, (1, 6))]}),
            (63, {"Jewels": (1, 4, 7500), "Magic_Items": [(magic_table_e, (1, 6))]}),
            (68, {"Gems": (1, 8, 5000), "Magic_Items": [(magic_table_e, (1, 6))]}),
            (69, {"Gems": (3, 6, 1000), "Magic_Items": [(magic_table_g, (1, 4))]}),
            (70, {"Jewels": (1, 10, 2500), "Magic_Items": [(magic_table_g, (1, 4))]}),
            (71, {"Jewels": (1, 4, 7500), "Magic_Items": [(magic_table_g, (1, 4))]}),
            (72, {"Gems": (1, 8, 5000), "Magic_Items": [(magic_table_g, (1, 4))]}),
            (74, {"Gems": (3, 6, 1000), "Magic_Items": [(magic_table_h, (1, 4))]}),
            (76, {"Jewels": (1, 10, 2500), "Magic_Items": [(magic_table_h, (1, 4))]}),
            (78, {"Jewels": (1, 4, 7500), "Magic_Items": [(magic_table_h, (1, 4))]}),
            (80, {"Gems": (1, 8, 5000), "Magic_Items": [(magic_table_h, (1, 4))]}),
            (85, {"Gems": (3, 6, 1000), "Magic_Items": [(magic_table_i, (1, 4))]}),
            (90, {"Jewels": (1, 10, 2500), "Magic_Items": [(magic_table_i, (1, 4))]}),
            (95, {"Jewels": (1, 4, 7500), "Magic_Items": [(magic_table_i, (1, 4))]}),
            (100, {"Gems": (1, 8, 5000), "Magic_Items": [(magic_table_i, (1, 4))]})
        )
        self.set_loot(loot_data)


magic_table_a = CumulativeWeightedChoice((
    (50, "Potion of healing"),
    (60, lambda: f"Spell scroll (cantrip) {random_spell('cantrip')}"),
    (70, "Potion of climbing"),
    (90, lambda: f"Spell scroll (1st level) {random_spell('level_1')}"),
    (94, lambda: f"Spell scroll (2nd level) {random_spell('level_2')}"),
    (98, "Potion of greater healing"),
    (99, "Bag of holding"),
    (100, "Driftglobe")
))


magic_table_b = CumulativeWeightedChoice((
    (15, "Potion of greater healing"),
    (22, "Potion of fire breath"),
    (29, "Potion of resistance"),
    (34, lambda: f"Ammunition: {random_ammunition()} +1"),
    (39, "Potion of animal friendship"),
    (44, "Potion of hill giant strength"),
    (49, "Potion of growth"),
    (54, "Potion of water breathing"),
    (59, lambda: f"Spell scroll (2nd level) {random_spell('level_2')}"),
    (64, lambda: f"Spell scroll (3rd level) {random_spell('level_3')}"),
    (67, "Bag of holding"),
    (70, "Keoghtom's ointment"),
    (73, "Oil of slipperiness"),
    (75, "Dust of disappearance"),
    (77, "Dust of dryness"),
    (79, "Dust of sneezing and choking"),
    (81, "Elemental gem"),
    (83, "Philter of love"),
    (84, "Alchemy jug"),
    (85, "Cap of water breathing"),
    (86, "Cloak of the manta ray"),
    (87, "Driftglobe"),
    (88, "Goggles of night"),
    (89, "Helm of comprehending languages"),
    (90, "Immovable rod"),
    (91, "Lantern of revealing"),
    (92, "Mariner's armor"),
    (93, "Mithral armor"),
    (94, "Potion of poison"),
    (95, "Ring of swimming"),
    (96, "Robe of useful items"),
    (97, "Rope of climbing"),
    (98, "Saddle of the cavalier"),
    (99, "Wand of magic detection"),
    (100, "Wand of secrets")
))


magic_table_c = CumulativeWeightedChoice((
    (15, "Potion of superior healing"),
    (22, lambda: f"Spell scroll (4th level) {random_spell('level_4')}"),
    (27, lambda: f"Ammunition: {random_ammunition()} +2"),
    (32, "Potion of clairvoyance"),
    (37, "Potion of diminution"),
    (42, "Potion of gaseous form"),
    (47, "Potion of frost giant strength"),
    (52, "Potion of stone giant strength"),
    (57, "Potion of heroism"),
    (62, "Potion of invulnerability"),
    (67, "Potion of mind reading"),
    (72, lambda: f"Spell scroll (5th level) {random_spell('level_5')}"),
    (75, "Elixir of health"),
    (78, "Oil of etherealness"),
    (81, "Potion of fire giant strength"),
    (84, "Quaal's feather token"),
    (87, "Scroll of protection"),
    (89, "Bag of beans"),
    (91, "Bead of force"),
    (92, "Chime of opening"),
    (93, "Decanter of endless water"),
    (94, "Eyes of minute seeing"),
    (95, "Folding boat"),
    (96, "Heward's handy haversack"),
    (97, "Horseshoes of speed"),
    (98, "Necklace of fireballs"),
    (99, "Periapt of health"),
    (100, "Sending stones")
))


magic_table_d = CumulativeWeightedChoice((
    (20, "Potion of supreme healing"),
    (30, "Potion of invisibility"),
    (40, "Potion of speed"),
    (50, lambda: f"Spell scroll (6th level) {random_spell('level_6')}"),
    (57, lambda: f"Spell scroll (7th level) {random_spell('level_7')}"),
    (62, lambda: f"Ammunition: {random_ammunition()} +3"),
    (67, "Oil of sharpness"),
    (72, "Potion of flying"),
    (77, "Potion of cloud giant strength"),
    (82, "Potion of longevity"),
    (87, "Potion of vitality"),
    (92, lambda: f"Spell scroll (8th level) {random_spell('level_8')}"),
    (95, "Horseshoes of a zephyr"),
    (98, "Nolzur's marvelous pigments"),
    (99, "Bag of devouring"),
    (100, "Portable hole")
))


magic_table_e = CumulativeWeightedChoice((
    (30, lambda: f"Spell scroll (8th level) {random_spell('level_8')}"),
    (55, "Potion of storm giant strength"),
    (70, "Potion of supreme healing"),
    (85, lambda: f"Spell scroll (9th level) {random_spell('level_9')}"),
    (93, "Universal solvent"),
    (98, "Arrow of slaying"),
    (100, "Sovereign glue")
))


magic_table_f = CumulativeWeightedChoice((
    (15, lambda: f"{weapon_list()} +1"),
    (18, "Shield +1"),
    (21, "Sentinel shield"),
    (23, "Amulet of proof against detection and location"),
    (25, "Boots of elvenkind"),
    (27, "Boots of striding and springing"),
    (29, "Bracers of archery"),
    (31, "Brooch of shielding"),
    (33, "Broom of flying"),
    (35, "Cloak of elvenkind"),
    (37, "Cloak of protection"),
    (39, "Gauntlets of ogre power"),
    (41, "Hat of disguise"),
    (43, "Javelin of lightning"),
    (45, "Pearl of power"),
    (47, "Rod of the pact keeper +1"),
    (49, "Slippers of spider climbing"),
    (51, "Staff of the adder"),
    (53, "Staff of the python"),
    (55, "Sword of vengeance"),
    (57, "Trident of fish command"),
    (59, "Wand of magic missiles"),
    (61, "Wand of the war mage +1"),
    (63, "Wand of web"),
    (65, "Weapon of warning"),
    (66, "Adamantine armor (chain mail)"),
    (67, "Adamantine armor (chain shirt)"),
    (68, "Adamantine armor (scale mail)"),
    (69, "Bag of tricks (gray)"),
    (70, "Bag of tricks (rust)"),
    (71, "Bag of tricks (tan)"),
    (72, "Boots of the winterlands"),
    (73, "Circlet of blasting"),
    (74, "Deck of illusions"),
    (75, "Eversmoking bottle"),
    (76, "Eyes of charming"),
    (77, "Eyes of the eagle"),
    (78, "Figurine of wondrous power (silver raven)"),
    (79, "Gem of brightness"),
    (80, "Gloves of missile snaring"),
    (81, "Gloves of swimming and climbing"),
    (82, "Gloves of thievery"),
    (83, "Headband of intellect"),
    (84, "Helm of telepathy"),
    (85, "Instrument of the bards (Doss lute)"),
    (86, "Instrument of the bards (Fochlucan bandore)"),
    (87, "Instrument of the bards (Mac-Fuimidh cittern)"),
    (88, "Medallion of thoughts"),
    (89, "Necklace of adaptation"),
    (90, "Periapt of wound closure"),
    (91, "Pipes of haunting"),
    (92, "Pipes of the sewers"),
    (93, "Ring of jumping"),
    (94, "Ring of mind shielding"),
    (95, "Ring of warmth"),
    (96, "Ring of water walking"),
    (97, "Quiver of Ehlonna"),
    (98, "Stone of good luck"),
    (99, "Wind fan"),
    (100, "Winged boots")
))


magic_table_g = CumulativeWeightedChoice((
    (11, lambda: f"{weapon_list()} +2"),
    (14, lambda: f"Figurine of wondrous power: {fig_wondrous_power()}"),
    (15, "Adamantine armor (breastplate)"),
    (16, "Adamantine armor (splint)"),
    (17, "Amulet of health"),
    (18, "Armor of vulnerability"),
    (19, "Arrow-catching shield"),
    (20, "Belt of dwarvenkind"),
    (21, "Belt of hill giant strength"),
    (22, "Berserker axe"),
    (23, "Boots of levitation"),
    (24, "Boots of speed"),
    (25, "Bowl of commanding water elemental"),
    (26, "Bracers of defense"),
    (27, "Brazier of commanding fire elemental"),
    (28, "Cape of the mountebank"),
    (29, "Censer of controlling air elemental"),
    (30, "Armor +1 chain mail"),
    (31, "Armor of resistance (chain mail)"),
    (32, "Armor +1 chain shirt"),
    (33, "Armor of resistance (chain shirt)"),
    (34, "Cloak of displacement"),
    (35, "Cloak of the bat"),
    (36, "Cube of force"),
    (37, "Daern's instant fortress"),
    (38, "Dagger of venom"),
    (39, "Dimensional shackles"),
    (40, "Dragon slayer"),
    (41, "Elven chain"),
    (42, "Flame tongue"),
    (43, "Gem of seeing"),
    (44, "Giant slayer"),
    (45, "Clamoured studded leather"),
    (46, "Helm of teleportation"),
    (47, "Horn of blasting"),
    (48, "Horn of Valhalla (silver or brass)"),
    (49, "Instrument of the bards (Canaith mandolin)"),
    (50, "Instrument of the bards (Cii lyre)"),
    (51, "Ioun stone (awareness)"),
    (52, "Ioun stone (protection)"),
    (53, "Ioun stone (reserve)"),
    (54, "Ioun stone (sustenance)"),
    (55, "Iron bands of Bilarro"),
    (56, "Armor +1 leather"),
    (57, "Armor of resistance (leather)"),
    (58, "Mace of disruption"),
    (59, "Mace of smiting"),
    (60, "Mace of terror"),
    (61, "Mantle of spell resistance"),
    (62, "Necklace of prayer beads"),
    (63, "Periapt of proof against poison"),
    (64, "Ring of animal influence"),
    (65, "Ring of evasion"),
    (66, "Ring of feather falling"),
    (67, "Ring of free action"),
    (68, "Ring of protection"),
    (69, "Ring of resistance"),
    (70, "Ring of spell storing"),
    (71, "Ring of the ram"),
    (72, "Ring of X-ray vision"),
    (73, "Robe of eyes"),
    (74, "Rod of rulership"),
    (75, "Rod of the pact keeper +2"),
    (76, "Rope of entanglement"),
    (77, "Armor +1 scale mail"),
    (78, "Armor of resistance (scale mail)"),
    (79, "Shield +2"),
    (80, "Shield of missile attraction"),
    (81, "Staff of charming"),
    (82, "Staff of healing"),
    (83, "Staff of swarming insects"),
    (84, "Staff of the woodlands"),
    (85, "Staff of withering"),
    (86, "Stone of controlling earth elemental"),
    (87, "Sun blade"),
    (88, "Sword of life stealing"),
    (89, "Sword of wounding"),
    (90, "Tentacle rod"),
    (91, "Vicious weapon"),
    (92, "Wand of binding"),
    (93, "Wand of enemy detection"),
    (94, "Wand of fear"),
    (95, "Wand of fireballs"),
    (96, "Wand of lightning bolts"),
    (97, "Wand of paralysis"),
    (98, "Wand of the war mage +2"),
    (99, "Wand of wonder"),
    (100, "Wings of flying")
))


magic_table_h = CumulativeWeightedChoice((
    (10, lambda: f"{weapon_list()} +3"),
    (12, "Amulet of the planes"),
    (14, "Carpet of flying"),
    (16, "Crystal ball (very rare version)"),
    (18, "Ring of regeneration"),
    (20, "Ring of shooting stars"),
    (22, "Ring of telekinesis"),
    (24, "Robe of scintillating colors"),
    (26, "Robe of stars"),
    (28, "Rod of absorption"),
    (30, "Rod of alertness"),
    (32, "Rod of security"),
    (34, "Rod of the pact keeper +3"),
    (36, "Scimitar of speed"),
    (38, "Shield +3"),
    (40, "Staff of fire"),
    (42, "Staff of frost"),
    (44, "Staff of power"),
    (46, "Staff of striking"),
    (48, "Staff of thunder and lightning"),
    (50, "Sword of sharpness"),
    (52, "Wand of polymorph"),
    (54, "Wand of the war mage +3"),
    (55, "Adamantine armor (half plate)"),
    (56, "Adamantine armor (plate)"),
    (57, "Animated shield"),
    (58, "Belt of fire giant strength"),
    (59, "Belt of frost (or stone) giant strength"),
    (60, "Armor +1 breastplate"),
    (61, "Armor of resistance (breastplate)"),
    (62, "Candle of invocation"),
    (63, "Armor +2 chain mail"),
    (64, "Armor +2 chain shirt"),
    (65, "Cloak of arachnida"),
    (66, "Dancing sword"),
    (67, "Demon armor"),
    (68, "Dragon scale mail"),
    (69, "Dwarven plate"),
    (70, "Dwarven thrower"),
    (71, "Efreeti bottle"),
    (72, "Figurine of wondrous power (obsidian steed)"),
    (73, "Frost brand"),
    (74, "Helm of brilliance"),
    (75, "Horn of Valhalla (bronze)"),
    (76, "Instrument of the bards (Anstruth harp)"),
    (77, "Ioun stone (absorption)"),
    (78, "Ioun stone (agility)"),
    (79, "Ioun stone (fortitude)"),
    (80, "Ioun stone (insight)"),
    (81, "Ioun stone (intellect)"),
    (82, "Ioun stone (leadership)"),
    (83, "Ioun stone (strength)"),
    (84, "Armor +2 leather"),
    (85, "Manual of bodily health"),
    (86, "Manual of gainful exercise"),
    (87, "Manual of golems"),
    (88, "Manual of quickness of action"),
    (89, "Mirror of life trapping"),
    (90, "Nine lives stealer"),
    (91, "Oath bow"),
    (92, "Armor +2 scale mail"),
    (93, "Spellguard shield"),
    (94, "Armor +1 splint"),
    (95, "Armor of resistance (splint)"),
    (96, "Armor +1 studded leather"),
    (97, "Armor of resistance (studded leather)"),
    (98, "Tome of clear thought"),
    (99, "Tome of leadership and influence"),
    (100, "Tome of understanding")
))


magic_table_i = CumulativeWeightedChoice((
    (5, "Defender"),
    (10, "Hammer of thunderbolts"),
    (15, "Luck blade"),
    (20, "Sword of answering"),
    (23, "Holy avenger"),
    (26, "Ring of djinni summoning"),
    (29, "Ring of invisibility"),
    (32, "Ring of spell turning"),
    (35, "Rod of lordly might"),
    (38, "Staff of the magi"),
    (41, "Vorpal sword"),
    (43, "Belt of cloud giant strength"),
    (45, "Armor +2 breastplate"),
    (47, "Armor +3 chain mail"),
    (49, "Armor +3 chain shirt"),
    (51, "Cloak of invisibility"),
    (53, "Crystal ball (legendary version)"),
    (55, "Armor +1 half plate"),
    (57, "Iron flask"),
    (59, "Armor +3 leather"),
    (61, "Armor +1 plate"),
    (63, "Robe of the archmagi"),
    (65, "Rod of resurrection"),
    (67, "Armor +1 scale mail"),
    (69, "Scarab of protection"),
    (71, "Armor +2 splint"),
    (73, "Armor +2 studded leather"),
    (75, "Well of many worlds"),
    (76, lambda: f"Armor {magic_armor()}"),
    (77, "Apparatus of Kwalish"),
    (78, "Armor of invulnerability"),
    (79, "Belt of storm giant strength"),
    (80, "Cubic gate"),
    (81, "Deck of many things"),
    (82, "Efreeti chain"),
    (83, "Armor of resistance (half plate)"),
    (84, "Horn of Valhalla (iron)"),
    (85, "Instrument of the bards (Ollamh harp)"),
    (86, "Ioun stone (greater absorption)"),
    (87, "Ioun stone (mastery)"),
    (88, "Ioun stone (regeneration)"),
    (89, "Plate armor of etherealness"),
    (90, "Plate armor of resistance"),
    (91, "Ring of air elemental command"),
    (92, "Ring of earth elemental command"),
    (93, "Ring of fire elemental command"),
    (94, "Ring of three wishes"),
    (95, "Ring of water elemental command"),
    (96, "Sphere of annihilation"),
    (97, "Talisman of pure good"),
    (98, "Talisman of the sphere"),
    (99, "Talisman of ultimate evil"),
    (100, "Tome of the stilled tongue")
))


weapon_list = TruffleShuffle((
    'Battleaxe',
    'Flail',
    'Glaive',
    'Greataxe',
    'Greatsword',
    'Halberd',
    'Lance',
    'Longsword',
    'Maul',
    'Morningstar',
    'Pike',
    'Rapier',
    'Scimitar',
    'Shortsword',
    'Trident',
    'War Pick',
    'Warhammer',
    'Whip',
    'Blowgun',
    'Hand Crossbow',
    'Heavy Crossbow',
    'Longbow',
    'Net',
    'Club',
    'Dagger',
    'Greatclub',
    'Handaxe',
    'Javelin',
    'Light Hammer',
    'Mace',
    'Quarterstaff',
    'Sickle',
    'Spear',
    'Light Crossbow',
    'Dart',
    'Shortbow',
    'Sling'
))

fig_wondrous_power = CumulativeWeightedChoice((
    (1, "Bronze griffon"),
    (2, "Ebony fly"),
    (3, "Golden lion"),
    (4, "Ivory goat"),
    (5, "Marble elephant"),
    (7, "Onyx dog"),
    (8, "Serpentine owl"),
))

magic_armor = CumulativeWeightedChoice((
    (2, "+2 half plate"),
    (4, "+2 plate"),
    (6, "+3 studded leather"),
    (8, "+3 breastplate"),
    (10, "+3 splint"),
    (11, "+3 half plate"),
    (12, "+3 plate"),
))


random_ammunition = TruffleShuffle((
    lambda: f"Quiver of {d(3) * 10} arrows",
    lambda: f"Box of {d(3) * 10} crossbow bolts",
    lambda: f"Bag of {d(3) * 10} sling stones",
))


def career_loot(party_size=5, start_level=1, end_level=20):
    total_loot = Loot()
    room_start = start_level * 10
    room_end = end_level * 10
    num_rooms = 0
    for i in range(room_start, room_end):
        num_rooms += 1
        total_loot += get_loot(
            cr_adapter(num_players=party_size, average_level=i//10),
            is_villain=percent_true(20)
        )
    return total_loot


def no_coin(loot: Loot) -> Loot:
    new_loot = Loot()
    new_loot.magic_items = loot.magic_items.copy()
    return new_loot


if __name__ == '__main__':
    print()
    print(RankFourHoardLoot())



from Fortuna import *


__all__ = (
    "Acolyte", "Charlatan", "Criminal", "Entertainer", "FolkHero", "GuildArtisan",
    "Hermit", "Noble", "Outlander", "Sage", "Sailor", "Soldier", "Urchin",
    "random_background",
)


class Background:
    name = ()
    feature = ()
    skills = ()
    proficiencies = ()
    equipment = ()
    trait = ()
    ideal = ()
    bond = ()
    flaw = ()

    def __str__(self):
        output = (
            f"{self.name}, {self.feature}",
            f"Ideal: {self.ideal}",
            f"Flaw: {self.flaw}",
        )
        return "\n".join(output)


class Acolyte(Background):
    name = "Acolyte"
    feature = "Shelter of the Faithful"
    skills = ("Insight WIS", "Religion INT")
    proficiencies = ("Choose two languages", )
    equipment = (
        "A holy symbol", "a prayer book or prayer wheel", "5 sticks of incense",
        "vestments", "a set of common clothes", "and a belt pouch containing 15 gp.",
    )
    random_trait = TruffleShuffle((
        "I idolize a particular hero of my faith, and constantly refer to that person's deeds and example.",
        "I see omens in every event and action.",
        "The gods try to speak to us, we just need to listen.",
        "Nothing can shake my optimistic attitude.",
        "I quote sacred texts and proverbs in almost every situation.",
        "I am tolerant of other faiths and respect the worship of other gods.",
        "I am intolerant of other faiths and condemn the worship of other gods.",
        "I've enjoyed fine food, drink, and high society among my temple's elite. Rough living grates on me.",
        "I've spent so long in the temple that I have little practical experience dealing with people in the outside world."
    ))
    random_ideal = TruffleShuffle((
        "Tradition. The ancient traditions of worship and sacrifice must be preserved and upheld.",
        "Charity. I always try to help those in need, no matter what the personal cost.",
        "Change. We must help bring about the changes the gods are constantly working in the world.",
        "Power. I hope to one day rise to the top of my faith's religious hierarchy.",
        "Faith. I trust that my deity will guide my actions, I have faith that if I work hard, things will go well.",
        "Aspiration. I seek to prove myself worthy of my god's favor by matching my actions against his or her teachings."
    ))
    random_bond = TruffleShuffle((
        "I would die to recover an ancient relic of my faith that was lost long ago.",
        "I will someday get revenge on the corrupt temple hierarchy who branded me a heretic.",
        "I owe my life to the priest who took me in when my parents died.",
        "Everything I do is for the common people.",
        "I will do anything to protect the temple where I served.",
        "I seek to preserve a sacred text that my enemies consider heretical and seek to destroy.",
    ))
    random_flaw = TruffleShuffle((
        "I judge others harshly, and myself even more severely.",
        "I put too much trust in those who wield power within my temple's hierarchy.",
        "My piety sometimes leads me to blindly trust those that profess faith in my god.",
        "I am inflexible in my thinking.",
        "I am suspicious of strangers and expect the worst of them.",
        "Once I pick a goal, I become obsessed with it to the detriment of everything else in my life.",
    ))

    def __init__(self):
        self.trait = self.random_trait()
        self.ideal = self.random_ideal()
        self.bond = self.random_bond()
        self.flaw = self.random_flaw()


class Charlatan(Background):
    name = "Charlatan"
    random_feature = TruffleShuffle((
        "Cheater",
        "Forgery",
        "Con-artist",
        "Disguises",
    ))
    skills = ("Deception CHA", "Sleight of Hand DEX")
    proficiencies = ("Disguise kit", "forgery kit")
    equipment = (
        "A set of fine clothes", "a disguise kit",
        "tools of the con of your choice (ten stoppered bottles filled with colored liquid",
        "a set of weighted dice", "a deck of marked cards", "or a signet ring of an imaginary duke)",
        "and a belt pouch containing 15 gp.")
    random_trait = TruffleShuffle((
        "I fall in and out of love easily, and am always pursuing someone.",
        "I have a joke for every occasion, especially occasions where humor is inappropriate.",
        "Flattery is my preferred trick for getting what I want.",
        "I'm a born gambler who can't resist taking a risk for a potential payoff.",
        "I lie about almost everything, even when there's no good reason to.",
        "Sarcasm and insults are my weapons of choice.",
        "I keep multiple holy symbols on me and invoke whatever deity might come in useful at any given moment.",
    ))
    random_ideal = TruffleShuffle((
        "Independence. I am a free spirit— no one tells me what to do.",
        "Fairness. I never target people who can't afford to lose a few coins.",
        "Charity. I distribute the money I acquire to the people who really need it.",
        "Creativity. I never run the same con twice.",
        "Friendship. Material goods come and go. Bonds of friendship last forever.",
        "Aspiration. I'm determined to make something of myself."
    ))
    random_bond = TruffleShuffle((
        "I fleeced the wrong person and must work to ensure that this individual never crosses paths with me or "
        "those I care about.",
        "I owe everything to my mentor—a horrible person who's probably rotting in jail somewhere.",
        "Somewhere out there, I have a child who doesn't know me. I'm making the world better for him or her.",
        "I come from a noble family, and one day I'll reclaim my lands and title from those who stole them from me.",
        "A powerful person killed someone I love. Some day soon, I'll have my revenge.",
        "I swindled and ruined a person who didn't deserve it. I seek to atone for my misdeeds but might never "
        "be able to forgive myself."
    ))
    random_flaw = TruffleShuffle((
        "I can't resist a pretty face.",
        "I'm always in debt. I spend my ill-gotten gains on decadent luxuries faster than I bring them in.",
        "I'm convinced that no one could ever fool me the way I fool others.",
        "I'm too greedy for my own good. I can't resist taking a risk if there's money involved.",
        "I can't resist swindling people who are more powerful than me.",
        "I hate to admit it and will hate myself for it, but I'll run and preserve my own hide if the going gets tough."
    ))

    def __init__(self):
        self.feature = self.random_feature()
        self.trait = self.random_trait()
        self.ideal = self.random_ideal()
        self.bond = self.random_bond()
        self.flaw = self.random_flaw()


class Criminal(Background):
    name = "Criminal"
    random_feature = TruffleShuffle((
        "Blackmailer",
        "Robber",
        "Burglar",
        "Assassin",
        "Enforcer",
        "Pickpocket",
        "Smuggler",
        "Spy",
    ))
    skills = ("Deception CHA", "Stealth DEX")
    proficiencies = ("One type of gaming set", "thieves' tools")
    equipment = ("A crowbar", "a set of dark common clothes including a hood", "and a belt pouch containing 15 gp.")
    random_trait = TruffleShuffle((
        "I always have a plan for what to do when things go wrong.",
        "I am always calm, no matter what the situation. I never raise my voice or let my emotions control me.",
        "The first thing I do in a new place is note the locations of everything of value--or where such things "
        "could be hidden",
        "I would rather make new friend than a new enemy.",
        "I am incredibly slow to trust. Those who seem the fairest often have the most to hide.",
        "I don't pay attention to the risks in a situation. Never tell me the odds.",
        "The best way to get me to do something is to tell me I can't do it.",
        "I blow up at the slightest insult.",
    ))
    random_ideal = TruffleShuffle((
        "Honor. I don't steal from others in the trade.",
        "Freedom. Chains are meant to be broken, as are those who would forge them.",
        "Charity. I steal from the wealthy so that I can help people in need.",
        "Greed. I will do whatever it takes to become wealthy.",
        "People. I'm loyal to my friends, not to any ideals, and everyone else can take a trip down the "
        "Styx for all I care.",
        "Redemption. There's a spark of good in everyone.",
    ))
    random_bond = TruffleShuffle((
        "I'm trying to pay off an old debt I owe to a generous benefactor.",
        "My ill-gotten gains go to support my family.",
        "Something important was taken from me, and I aim to steal it back.",
        "I will become the greatest thief that ever lived.",
        "I'm guilty of a terrible crime. I hope I can redeem myself for it.",
        "Someone I loved died because of a mistake I made. That will never happen again.",
    ))
    random_flaw = TruffleShuffle((
        "When I see something valuable, I can't think about anything but how to steal it.",
        "When faced with a choice between money and my friends, I usually choose the money.",
        "If there's a plan, I'll forget it. If I don't forget it, I'll ignore it.",
        "I have a 'tell' that reveals when I'm lying.",
        "I turn tail and run when things look bad.",
        "An innocent person is in prison for a crime that I committed. I'm okay with that.",
    ))

    def __init__(self):
        self.feature = self.random_feature()
        self.trait = self.random_trait()
        self.ideal = self.random_ideal()
        self.bond = self.random_bond()
        self.flaw = self.random_flaw()


class Entertainer(Background):
    name = "Entertainer"
    random_feature = TruffleShuffle((
        "Actor",
        "Dancer",
        "Fire-eater",
        "Jester",
        "Juggler",
        "Instrumentalist",
        "Poet",
        "Storyteller",
        "Tumbler",
        "Gladiator",
    ))
    skills = ("Acrobatics DEX", "Performance CHA")
    proficiencies = ("Disguise kit", "one type of musical instrument")
    equipment = (
        "A musical instrument (one of your choice)",
        "the favor of an admirer (love letter, lock of hair, or trinket)",
        "a costume", "and a belt pouch containing 15 gp."
    )
    random_trait = TruffleShuffle((
        "I know a story relevant to almost every situation.",
        "Whenever I come to a new place, I collect local rumors and spread gossip.",
        "I'm a hopeless romantic, always searching for that special someone.",
        "Nobody stays angry at me or around me for long, since I can defuse any amount of tension.",
        "I love a good insult, even one directed at me.",
        "I get bitter if I'm not the center of attention.",
        "I'll settle for nothing less than perfection.",
        "I change my mood or my mind as quickly as I change key in a song.",
    ))
    random_ideal = TruffleShuffle((
        "Beauty. When I perform, I make the world better than it was.",
        "Tradition. The stories, legends, and songs of the past must never be forgotten, for they teach us who we are.",
        "Creativity. The world is in need of new ideas and bold action.",
        "Greed. I'm only in it for the money and fame.",
        "People. I like seeing the smiles on people's faces when I perform. That's all that matters.",
        "Honesty. Art should reflect the soul; it should come from within and reveal who we really are.",
    ))
    random_bond = TruffleShuffle((
        "My instrument is my most treasured possession, and it reminds me of someone I love.",
        "Someone stole my precious instrument, and someday I'll get it back.",
        "I want to be famous, whatever it takes.",
        "I idolize a hero of the old tales and measure my deeds against that person's.",
        "I will do anything to prove myself superior to my hated rival.",
        "I would do anything for the other members of my old troupe.",
    ))
    random_flaw = TruffleShuffle((
        "I'll do anything to win fame and renown.",
        "I'm a sucker for a pretty face.",
        "A scandal prevents me from ever going home again. That kind of trouble seems to follow me around.",
        "I once satirized a noble who still wants my head. It was a mistake that I will likely repeat.",
        "I have trouble keeping my true feelings hidden. My sharp tongue lands me in trouble.",
        "Despite my best efforts, I am unreliable to my friends.",
    ))

    def __init__(self):
        self.feature = self.random_feature()
        self.trait = self.random_trait()
        self.ideal = self.random_ideal()
        self.bond = self.random_bond()
        self.flaw = self.random_flaw()


class FolkHero(Background):
    name = "Folk Hero"
    feature = "Hero of the People"
    skills = ("Animal Handling WIS", "Survival WIS")
    proficiencies = ("One type of artisan's tools", "vehicles (land)")
    equipment = (
        "A set of artisan's tools (one of your choice)",
        "a shovel", "an iron pot", "a set of common clothes",
        "and a belt pouch containing 10 gp.",
    )
    random_trait = TruffleShuffle((
        "I judge people by their actions, not their words.",
        "If someone is in trouble, I'm always ready to lend help.",
        "When I set my mind to something, I follow through no matter what gets in my way.",
        "I have a strong sense of fair play and always try to find the most equitable solution to arguments.",
        "I'm confident in my own abilities and do what I can to instill confidence in others.",
        "Thinking is for other people. I prefer action.",
        "I misuse long words in an attempt to sound smarter.",
        "I get bored easily. When am I going to get on with my destiny?",
    ))
    random_ideal = TruffleShuffle((
        "Respect. People deserve to be treated with dignity and respect.",
        "Fairness. No one should get preferential treatment before the law, and no one is above the law.",
        "Freedom. Tyrants must not be allowed to oppress the people.",
        "Might. If I become strong, I can take what I want—-what I deserve.",
        "Sincerity. There's no good in pretending to be something I'm not.",
        "Destiny. Nothing and no one can steer me away from my higher calling.",
    ))
    random_bond = TruffleShuffle((
        "I have a family, but I have no idea where they are. One day, I hope to see them again.",
        "I worked the land, I love the land, and I will protect the land.",
        "A proud noble once gave me a horrible beating, and I will take my revenge on any bully I encounter.",
        "My tools are symbols of my past life, and I carry them so that I will never forget my roots.",
        "I protect those who cannot protect themselves.",
        "I wish my childhood sweetheart had come with me to pursue my destiny.",
    ))
    random_flaw = TruffleShuffle((
        "The tyrant who rules my land will stop at nothing to see me killed.",
        "I'm convinced of the significance of my destiny, and blind to my shortcomings and the risk of failure.",
        "The people who knew me when I was young know my shameful secret, so I can never go home again.",
        "I have a weakness for the vices of the city, especially hard drink.",
        "Secretly, I believe that things would be better if I were a tyrant lording over the land.",
        "I have trouble trusting in my allies.",
    ))

    def __init__(self):
        self.trait = self.random_trait()
        self.ideal = self.random_ideal()
        self.bond = self.random_bond()
        self.flaw = self.random_flaw()


class GuildArtisan(Background):
    name = "Guild Artisan"
    random_feature = TruffleShuffle((
        "Alchemist",
        "Armorer",
        "Brewer",
        "Calligrapher",
        "Carpenter",
        "Cartographer",
        "Cobbler",
        "Cook",
        "Glassblowers",
        "Jeweler",
        "Leather Worker",
        "Mason",
        "Painter",
        "Potter",
        "Shipwright",
        "Blacksmith",
        "Tinker",
        "Wheelwright",
        "Weaver",
    ))
    skills = ("Insight WIS", "Persuasion CHA")
    proficiencies = ("One type of artisan's tools", "and one language of your choice")
    equipment = (
        "A set of artisan's tools (one of your choice)",
        "a letter of introduction from your guild", "a set of traveler's clothes",
        "and a belt pouch containing 15 gp."
    )
    random_trait = TruffleShuffle((
        "I believe that anything worth doing is worth doing right. I can't help it— I'm a perfectionist.",
        "I'm a snob who looks down on those who can't appreciate fine art.",
        "I always want to know how things work and what makes people tick.",
        "I'm full of witty aphorisms and have a proverb for every occasion.",
        "I'm rude to people who lack my commitment to hard work and fair play.",
        "I like to talk at length about my profession.",
        "I don't part with my money easily and will haggle tirelessly to get the best deal possible.",
        "I'm well known for my work, and I want to make sure everyone appreciates it. "
        "I'm always taken aback when people haven't heard of me.",
    ))
    random_ideal = TruffleShuffle((
        "Community. It is the duty of all civilized people to strengthen the bonds of community and the "
        "security of civilization.",
        "Generosity. My talents were given to me so that I could use them to benefit the world.",
        "Freedom. Everyone should be free to pursue his or her own livelihood.",
        "Greed. I'm only in it for the money.",
        "People. I'm committed to the people I care about, not to ideals.",
        "Aspiration. I work hard to be the best there is at my craft.",
    ))
    random_bond = TruffleShuffle((
        "The workshop where I learned my trade is the most important place in the world to me.",
        "I created a great work for someone, and then found them unworthy to receive it. I'm still looking for "
        "someone worthy.",
        "I owe my guild a great debt for forging me into the person I am today.",
        "I pursue wealth to secure someone's love.",
        "One day I will return to my guild and prove that I am the greatest artisan of them all.",
        "I will get revenge on the evil forces that destroyed my business.",
    ))
    random_flaw = TruffleShuffle((
        "I'll do anything to get my hands on something rare or priceless.",
        "I'm quick to assume that someone is trying to cheat me.",
        "No one must ever learn that I once stole money from guild coffers.",
        "I'm never satisfied with what I have—I always want more.",
        "I would kill to acquire a noble title.",
        "I'm horribly jealous of anyone who can outshine my handiwork.",
    ))

    def __init__(self):
        self.feature = self.random_feature()
        self.trait = self.random_trait()
        self.ideal = self.random_ideal()
        self.bond = self.random_bond()
        self.flaw = self.random_flaw()


class Hermit(Background):
    name = "Hermit"
    random_feature = TruffleShuffle((
        "Spiritual Enlightenment",
        "Religious Mandate",
        "Exiled Criminal",
        "Retired",
        "Secluded Artist",
        "Communing with Nature",
        "Caretaker of a Relic",
    ))
    skills = ("Medicine WIS", "Religion INT")
    proficiencies = ("Herbalism kit", "and one language of your choice.")
    equipment = (
        "A scroll case stuffed full of notes from your studies or prayers",
        "a winter blanket", "a set of common clothes", "an herbalism kit", "and 5 gp."
    )
    random_trait = TruffleShuffle((
        "I've been isolated for so long that I rarely speak preferring gestures and the occasional grunt.",
        "I am utterly serene, even in the face of disaster.",
        "The leader of my community had something wise to say on every topic, and I am eager to share that wisdom.",
        "I feel tremendous empathy for all who suffer.",
        "I'm oblivious to etiquette and social expectations.",
        "I connect everything that happens to me to a grand, cosmic plan.",
        "I often get lost in my own thoughts and contemplation, becoming oblivious to my surroundings.",
        "I am working on a grand philosophical theory and love sharing my ideas.",
    ))
    random_ideal = TruffleShuffle((
        "Greater Good. My gifts are meant to be shared with all, not used for my own benefit.",
        "Logic. Emotions must not cloud our sense of what is right and true, or our logical thinking.",
        "Free Thinking. Inquiry and curiosity are the pillars of progress.",
        "Power. Solitude and contemplation are paths toward mystical or magical power.",
        "Live and Let Live. Meddling in the affairs of others only causes trouble.",
        "Self-Knowledge. If you know yourself, there's nothing left to know.",
    ))
    random_bond = TruffleShuffle((
        "Nothing is more important than the other members of my hermitage, order, or association.",
        "I entered seclusion to hide from the ones who might still be hunting me. I must someday confront them.",
        "I'm still seeking the enlightenment I pursued in my seclusion, and it still eludes me.",
        "I entered seclusion because I loved someone I could not have.",
        "Should my discovery come to light, it could bring ruin to the world.",
        "My isolation gave me",
    ))
    random_flaw = TruffleShuffle((
        "Now that I've returned to the world, I enjoy its delights a little too much.",
        "I harbor dark, bloodthirsty thoughts that my isolation and meditation failed to quell.",
        "I am dogmatic in my thoughts and philosophy.",
        "I let my need to win arguments overshadow friendships and harmony.",
        "I'd risk too much to uncover a lost bit of knowledge.",
        "I like keeping secrets and won't share them with anyone.",
    ))

    def __init__(self):
        self.feature = self.random_feature()
        self.trait = self.random_trait()
        self.ideal = self.random_ideal()
        self.bond = self.random_bond()
        self.flaw = self.random_flaw()


class Noble(Background):
    name = "Noble"
    feature = "Position of Privilege"
    skills = ("History INT", "Persuasion CHA")
    proficiencies = ("One type of gaming set", "and one language of your choice")
    equipment = ("A set of fine clothes", "a signet ring", "a scroll of pedigree", "and a purse containing 25 gp.")
    random_trait = TruffleShuffle((
        "My eloquent flattery makes everyone I talk to feel like the most wonderful and important person in the world.",
        "The common folk love me for my kindness and generosity.",
        "No one could doubt by looking at my regal bearing that I am a cut above the unwashed masses.",
        "I take great pains to always look my best and follow the latest fashions.",
        "I don't like to get my hands dirty, and I won't be caught dead in unsuitable accommodations.",
        "Despite my noble birth, I do not place myself above other folk. We all have the same blood.",
        "My favor, once lost, is lost forever.",
        "If you do me an injury, I will crush you, ruin your name, and salt your fields.",
    ))
    random_ideal = TruffleShuffle((
        "Respect. Respect is due to me because of my position, "
        "but all people regardless of station deserve to be treated with dignity.",
        "Responsibility. It is my duty to respect the authority of those above me, "
        "just as those below me must respect mine.",
        "Independence. I must prove that I can handle myself without the coddling of my family.",
        "Power. If I can attain more power, no one will tell me what to do.",
        "Family. Blood runs thicker than water.",
        "Noble Obligation. It is my duty to protect and care for the people beneath me.",
    ))
    random_bond = TruffleShuffle((
        "I will face any challenge to win the approval of my family.",
        "My house's alliance with another noble family must be sustained at all costs.",
        "Nothing is more important than the other members of my family.",
        "I am in love with the heir of a family that my family despises.",
        "My loyalty to my sovereign is unwavering.",
        "The common folk must see me as a hero of the people.",
    ))
    random_flaw = TruffleShuffle((
        "I secretly believe that everyone is beneath me.",
        "I hide a truly scandalous secret that could ruin my family forever.",
        "I too often hear veiled insults and threats in every word addressed to me, and I'm quick to anger.",
        "I have an insatiable desire for carnal pleasures.",
        "In fact, the world does revolve around me.",
        "By my words and actions, I often bring shame to my family.",
    ))

    def __init__(self):
        self.trait = self.random_trait()
        self.ideal = self.random_ideal()
        self.bond = self.random_bond()
        self.flaw = self.random_flaw()


class Outlander(Background):
    name = "Outlander"
    random_feature = TruffleShuffle((
        "Forester",
        "Trapper",
        "Homesteader",
        "Wilderness Guide",
        "Outcast",
        "Bounty Hunter",
        "Pilgrim",
        "Tribal Nomad",
        "Tribal Marauder",
    ))
    skills = ("Athletics STR", "Survival WIS")
    proficiencies = ("One type of musical instrument", "and one language of your choice")
    equipment = (
        "A staff", "a hunting trap", "a trophy from an animal you killed",
        "a set of traveler's clothes", "and a belt pouch containing 10 gp."
    )
    random_trait = TruffleShuffle((
        "I'm driven by a wanderlust that led me away from home.",
        "I watch over my friends as if they were a litter of newborn pups.",
        "I once ran twenty-five miles without stopping to warn to my clan of an approaching orc horde. "
        "I'd do it again if I had to.",
        "I have a lesson for every situation, drawn from observing nature.",
        "I place no stock in wealthy or well-mannered folk. Money and manners won't save you from a hungry owlbear.",
        "I'm always picking things up, absently fiddling with them, and sometimes accidentally breaking them.",
        "I feel far more comfortable around animals than people.",
        "I was, in fact, raised by wolves.",
    ))
    random_ideal = TruffleShuffle((
        "Change. Life is like the seasons, in constant change, and we must change with it.",
        "Greater Good. It is each person's responsibility to make the most happiness for the whole tribe.",
        "Honor. If I dishonor myself, I dishonor my whole clan.",
        "Might. The strongest are meant to rule.",
        "Nature. The natural world is more important than all the constructs of civilization.",
        "Glory. I must earn glory in battle, for myself and my clan.",
    ))
    random_bond = TruffleShuffle((
        "My family, clan, or tribe is the most important thing in my life, even when they are far from me.",
        "An injury to the unspoiled wilderness of my home is an injury to me.",
        "I will bring terrible wrath down on the evildoers who destroyed my homeland.",
        "I am the last of my tribe, and it is up to me to ensure their names enter legend.",
        "I suffer awful visions of a coming disaster and will do anything to prevent it.",
        "It is my duty to provide children to sustain my tribe.",
    ))
    random_flaw = TruffleShuffle((
        "I am too enamored of ale, wine, and other intoxicants.",
        "There's no room for caution in a life lived to the fullest.",
        "I remember every insult I've received and nurse a silent resentment toward anyone who's ever wronged me.",
        "I am slow to trust members of other races, tribes, and societies.",
        "Violence is my answer to almost any challenge.",
        "Don't expect me to save those who can't save themselves. It is nature's way that the strong thrive "
        "and the weak perish.",
    ))

    def __init__(self):
        self.feature = self.random_feature()
        self.trait = self.random_trait()
        self.ideal = self.random_ideal()
        self.bond = self.random_bond()
        self.flaw = self.random_flaw()


class Sage(Background):
    name = "Sage"
    random_feature = TruffleShuffle((
        "Alchemist",
        "Astronomer",
        "Librarian",
        "Professor",
        "Arcane Lore Master",
        "Wizard's Apprentice",
        "Scribe",
    ))
    skills = ("Arcana INT", "History INT")
    proficiencies = ("Two languages of your choice",)
    equipment = (
        "A bottle of black ink, a quill, a small knife",
        "a letter from a dead colleague posing a question you have not yet been able to answer",
        "a set of common clothes", "and a belt pouch containing 10 gp."
    )
    random_trait = TruffleShuffle((
        "I use polysyllabic words that convey the impression of great erudition.",
        "I've read every book in the world's greatest libraries—or I like to boast that I have.",
        "I'm used to helping out those who aren't as smart as I am, "
        "and I patiently explain anything and everything to others.",
        "There's nothing I like more than a good mystery.",
        "I'm willing to listen to every side of an argument before I make my own judgment.",
        "I... speak... slowly... when talking... to idiots... which... almost... "
        "everyone... is... compared... to me.",
        "I am horribly, horribly awkward in social situations.",
        "I'm convinced that people are always trying to steal my secrets.",
    ))
    random_ideal = TruffleShuffle((
        "Knowledge. The path to power and self-improvement is through knowledge.",
        "Beauty. What is beautiful points us beyond itself toward what is true.",
        "Logic. Emotions must not cloud our logical thinking.",
        "No Limits. Nothing should fetter the infinite possibility inherent in all existence.",
        "Power. Knowledge is the path to power and domination.",
        "Self-Improvement. The goal of a life of study is the betterment of oneself.",
    ))
    random_bond = TruffleShuffle((
        "It is my duty to protect my students.",
        "I have an ancient text that holds terrible secrets that must not fall into the wrong hands.",
        "I work to preserve a library, university, scriptorium, or monastery.",
        "My life's work is a series of tomes related to a specific field of lore.",
        "I've been searching my whole life for the answer to a certain question.",
        "I sold my soul for knowledge. I hope to do great deeds and win it back.",
    ))
    random_flaw = TruffleShuffle((
        "I am easily distracted by the promise of information.",
        "Most people scream and run when they see a demon. I stop and take notes on its anatomy.",
        "Unlocking an ancient mystery is worth the price of a civilization.",
        "I overlook obvious solutions in favor of complicated ones.",
        "I speak without really thinking through my words, invariably insulting others.",
        "I can't keep a secret to save my life, or anyone else's.",
    ))

    def __init__(self):
        self.feature = self.random_feature()
        self.trait = self.random_trait()
        self.ideal = self.random_ideal()
        self.bond = self.random_bond()
        self.flaw = self.random_flaw()


class Sailor(Background):
    name = "Sailor"
    feature = "Ship's Passage"
    skills = ("Athletics STR", "Perception WIS")
    proficiencies = ("Navigator's tools", "vehicles (water)")
    equipment = (
        "A belaying pin (club), 50 feet of silk rope",
        "a lucky charm such as a rabbit foot or a small stone with a hole in the center",
        "a set of common clothes", "and a belt pouch containing 10 gp."
    )
    random_trait = TruffleShuffle((
        "My friends know they can rely on me, no matter what.",
        "I work hard so that I can play hard when the work is done.",
        "I enjoy sailing into new ports and making new friends over a flagon of ale.",
        "I stretch the truth for the sake of a good story.",
        "To me, a tavern brawl is a nice way to get to know a new city.",
        "I never pass up a friendly wager.",
        "My language is as foul as an otyugh nest.",
        "I like a job well done, especially if I can convince someone else to do it.",
    ))
    random_ideal = TruffleShuffle((
        "Respect. The thing that keeps a ship together is mutual respect between captain and crew.",
        "Fairness. We all do the work, so we all share in the rewards.",
        "Freedom. The sea is freedom—the freedom to go anywhere and do anything.",
        "Mastery. I'm a predator, and the other ships on the sea are my prey.",
        "People. I'm committed to my crewmates, not to ideals.",
        "Aspiration. Someday I'll own my own ship and chart my own destiny.",
    ))
    random_bond = TruffleShuffle((
        "I'm loyal to my captain first, everything else second.",
        "The ship is most important—crewmates and captains come and go.",
        "I'll always remember my first ship.",
        "In a harbor town, I have a paramour whose eyes nearly stole me from the sea.",
        "I was cheated out of my fair share of the profits, and I want to get my due.",
        "Ruthless pirates murdered my captain and crewmates, plundered our ship, and left me to die. "
        "Vengeance will be mine.",
    ))
    random_flaw = TruffleShuffle((
        "I follow orders, even if I think they're wrong.",
        "I'll say anything to avoid having to do extra work.",
        "Once someone questions my courage, I never back down no matter how dangerous the situation.",
        "Once I start drinking, it's hard for me to stop.",
        "I can't help but pocket loose coins and other trinkets I come across.",
        "My pride will probably lead to my destruction.",
    ))

    def __init__(self):
        self.trait = self.random_trait()
        self.ideal = self.random_ideal()
        self.bond = self.random_bond()
        self.flaw = self.random_flaw()


class Soldier(Background):
    name = "Soldier"
    random_feature = TruffleShuffle((
        "Officer",
        "Scout",
        "Infantry",
        "Cavalry",
        "Healer",
        "Quartermaster",
        "Standard Bearer",
    ))
    skills = ("Athletics STR", "Intimidation CHA")
    proficiencies = ("One type of gaming set", "vehicles (land)")
    equipment = (
        "An insignia of rank, a trophy taken from a fallen enemy (a dagger, broken blade, or piece of a banner)",
        "a set of bone dice or deck of cards", "a set of common clothes", "and a belt pouch containing 10 gp.",
    )
    random_trait = TruffleShuffle((
        "I'm always polite and respectful.",
        "I'm haunted by memories of war. I can't get the images of violence out of my mind.",
        "I've lost too many friends, and I'm slow to make new ones.",
        "I'm full of inspiring and cautionary tales from my military experience "
        "relevant to almost every combat situation.",
        "I can stare down a hell hound without flinching.",
        "I enjoy being strong and like breaking things.",
        "I have a crude sense of humor.",
        "I face problems head-on. A simple, direct solution is the best path to success.",
    ))
    random_ideal = TruffleShuffle((
        "Greater Good. Our lot is to lay down our lives in defense of others.",
        "Responsibility. I do what I must and obey just authority.",
        "Independence. When people follow orders blindly, they embrace a kind of tyranny.",
        "Might. In life as in war, the stronger force wins.",
        "Live and Let Live. Ideals aren't worth killing over or going to war for.",
        "Nation. My city, nation, or people are all that matter.",
    ))
    random_bond = TruffleShuffle((
        "I would still lay down my life for the people I served with.",
        "Someone saved my life on the battlefield. To this day, I will never leave a friend behind.",
        "My honor is my life.",
        "I'll never forget the crushing defeat my company suffered or the enemies who dealt it.",
        "Those who fight beside me are those worth dying for.",
        "I fight for those who cannot fight for themselves.",
    ))
    random_flaw = TruffleShuffle((
        "The monstrous enemy we faced in battle still leaves me quivering with fear.",
        "I have little respect for anyone who is not a proven warrior.",
        "I made a terrible mistake in battle, it cost many lives and I would do anything to keep that mistake secret.",
        "My hatred of my enemies is blind and unreasoning.",
        "I obey the law, even if the law causes misery.",
        "I'd rather eat my armor than admit when I'm wrong.",
    ))

    def __init__(self):
        self.feature = self.random_feature()
        self.trait = self.random_trait()
        self.ideal = self.random_ideal()
        self.bond = self.random_bond()
        self.flaw = self.random_flaw()


class Urchin(Background):
    name = "Urchin"
    feature = "City Secrets"
    skills = ("Sleight of Hand DEX", "Stealth DEX")
    proficiencies = ("Disguise kit", "thieves' tools")
    equipment = (
        "A small knife, a map of the city you grew up in", "a pet mouse",
        "a token to remember your parents by", "a set of common clothes",
        "and a belt pouch containing 10 gp."
    )
    random_trait = TruffleShuffle((
        "I hide scraps of food and trinkets away in my pockets.",
        "I ask a lot of questions.",
        "I like to squeeze into small places where no one else can get to me.",
        "I sleep with my back to a wall or tree, with everything I own wrapped in a bundle in my arms.",
        "I eat like a pig and have bad manners.",
        "I think anyone who's nice to me is hiding evil intent.",
        "I don't like to bathe.",
        "I bluntly say what other people are hinting at or hiding.",
    ))
    random_ideal = TruffleShuffle((
        "Respect. All people, rich or poor, deserve respect.",
        "Community. We have to take care of each other, because no one else is going to do it.",
        "Change. The low are lifted up, and the high and mighty are brought down. Change is the nature of things.",
        "Retribution. The rich need to be shown what life and death are like in the gutters.",
        "People. I help the people who help me— that's what keeps us alive.",
        "Aspiration. I'm going to prove that I'm worthy of a better life.",
    ))
    random_bond = TruffleShuffle((
        "My town or city is my home, and I'll fight to defend it.",
        "I sponsor an orphanage to keep others from enduring what I was forced to endure.",
        "I owe my survival to another urchin who taught me to live on the streets.",
        "I owe a debt I can never repay to the person who took pity on me.",
        "I escaped my life of poverty by robbing an important person, and I'm wanted for it.",
        "No one else should have to endure the hardships I've been through.",
    ))
    random_flaw = TruffleShuffle((
        "If I'm outnumbered, I will run away from a fight.",
        "Gold seems like a lot of money to me, and I'll do just about anything for more of it.",
        "I will never fully trust anyone other than myself.",
        "I'd rather kill someone in their sleep then fight fair.",
        "It's not stealing if I need it more than someone else.",
        "People who can't take care of themselves get what they deserve.",
    ))

    def __init__(self):
        self.trait = self.random_trait()
        self.ideal = self.random_ideal()
        self.bond = self.random_bond()
        self.flaw = self.random_flaw()


backgrounds = (
    Acolyte,
    Charlatan,
    Criminal,
    Entertainer,
    FolkHero,
    GuildArtisan,
    Hermit,
    Noble,
    Outlander,
    Sage,
    Sailor,
    Soldier,
    Urchin,
)

random_background = TruffleShuffle(backgrounds)


if __name__ == "__main__":
    print()
    for background in backgrounds:
        print(background(), '\n')



from Fortuna import *

from DungeonGenerator.backgrounds import random_background
from DungeonGenerator.gear_lib import *
from DungeonGenerator.monsters_lib import damage_type_cat
from DungeonGenerator.treasure import Loot, get_loot, magic_table_by_cr
from DungeonGenerator.utilities import damage_types


class Base:
    damage_dice = 8

    def damage_str(self, num_dam_dice, bonus):
        return f"{num_dam_dice}d{self.damage_dice} + {bonus}"

    def damage(self, num_dam_dice, bonus):
        return dice(num_dam_dice, self.damage_dice) + bonus


class Apprentice(Base):
    hd = 4
    stats = {
        'STR': ability_dice(3),
        'DEX': ability_dice(3),
        'CON': ability_dice(3),
        'INT': ability_dice(3),
        'WIS': ability_dice(3),
        'CHA': ability_dice(3),
    }
    random_ability = TruffleShuffle(set(stats.keys()))
    fav_stat = max(stats)
    sec_stat = "CON"
    armor_ac = 10
    max_ac_dex = 4
    random_sub_class = TruffleShuffle((
        "Acolyte",
        "Scribe",
        "Squire",
        "Cultist",
        "Neophyte",
    ))

    def __init__(self):
        self.save_prof = {self.fav_stat, self.random_ability()}
        self.sub_class = self.random_sub_class()
        self.name = f"Apprentice, {self.sub_class}"
        self.damage_type = f"{damage_types('Physical')}"
        self.pref_weapon = villager_weapons()


class Barbarian(Base):
    hd = 12
    fav_stat = 'STR'
    sec_stat = 'DEX'
    primary_att = 'Melee Weapon'
    armor_ac = 10
    max_ac_dex = 4
    random_sub_class = TruffleShuffle((
        "Berserker",
        TruffleShuffle(("Bear Totem", "Eagle Totem", "Wolf Totem")),
        "Ancestral Guardian",
        "Storm Herald",
        "Zealot",
    ))
    stats = {'STR': 15, 'DEX': 13, 'CON': 14, 'INT': 8, 'WIS': 12, 'CHA': 10}
    save_prof = ('STR', 'CON')

    def __init__(self):
        self.pref_weapon = two_hand_weapons()
        self.sub_class = self.random_sub_class()
        self.name = f"Barbarian, Path of the {self.sub_class}"
        if self.sub_class == "Storm Herald":
            self.damage_type = f"{damage_types('Physical')} or lightning"
        if self.sub_class == "Ancestral Guardian" or self.sub_class == "Zealot":
            self.damage_type = f"{damage_types('Physical')} and holy fire"
        else:
            self.damage_type = f"{damage_types('Physical')}"


class Bard(Base):
    hd = 8
    fav_stat = 'CHA'
    sec_stat = 'DEX'
    primary_att = 'Charm'
    armor_ac = 12
    max_ac_dex = 4
    random_sub_class = TruffleShuffle((
        "Lore",
        "Valor",
        "Glamour",
        "Swords",
        "Whispers",
    ))
    stats = {'STR': 8, 'DEX': 14, 'CON': 13, 'INT': 10, 'WIS': 12, 'CHA': 15}
    save_prof = ('DEX', 'CHA')

    def __init__(self):
        self.sub_class = self.random_sub_class()
        self.name = f"Bard, College of {self.sub_class}"
        if self.sub_class == "Valor":
            self.damage_type = f"{damage_types('Physical')}"
            self.pref_weapon = main_hand_weapons()
        if self.sub_class == "Swords":
            self.damage_type = f"{damage_types('Physical')}"
            self.pref_weapon = "Longsword" if percent_true(50) else "Rapier"
        else:
            self.damage_type = f"{damage_types('Arcane')}"
            self.pref_weapon = caster_weapons()


class Cleric(Base):
    hd = 8
    name = 'Cleric'
    fav_stat = 'WIS'
    sec_stat = 'STR'
    primary_att = 'Spell'
    armor_ac = 15
    max_ac_dex = 2
    random_sub_class = TruffleShuffle((
        "Knowledge",
        "Life",
        "Light",
        "Nature",
        "Tempest",
        "Trickery",
        "War",
        "Forge",
        "Grave",
        "Death",
    ))
    stats = {'STR': 8, 'DEX': 12, 'CON': 14, 'INT': 10, 'WIS': 15, 'CHA': 13}
    save_prof = ('WIS', 'CHA')

    def __init__(self):
        self.sub_class = self.random_sub_class()
        self.name = f"Cleric, {self.sub_class} Domain"
        if self.sub_class == "War" or self.sub_class == "Forge":
            self.damage_type = f"{damage_types('Physical')} and {damage_types('Holy')}"
            self.pref_weapon = f"{off_hand_shield()} and {main_hand_weapons()}"
        elif self.sub_class == "Nature" or self.sub_class == "Tempest":
            self.damage_type = damage_types('Nature')
            self.pref_weapon = caster_weapons()
        elif self.sub_class == "Death" or self.sub_class == "Grave":
            self.damage_type = damage_types('Unholy')
            self.pref_weapon = caster_weapons()
        elif self.sub_class == "Light" or self.sub_class == "Life":
            self.damage_type = damage_types('Holy')
            self.pref_weapon = caster_weapons() if percent_true(75) else villager_weapons()
        else:
            self.damage_type = f"{damage_types('Divine')}"
            self.pref_weapon = f"{off_hand_weapons()} and {off_hand_shield()}"


class Druid(Base):
    hd = 8
    fav_stat = 'WIS'
    sec_stat = 'DEX'
    primary_att = 'Nature Spell'
    armor_ac = 15
    max_ac_dex = 2
    random_sub_class = TruffleShuffle((
        "the Land",
        "the Moon",
        "Dreams",
        "the Shepherd",
        "Spores",
    ))
    stats = {'STR': 13, 'DEX': 8, 'CON': 14, 'INT': 10, 'WIS': 15, 'CHA': 12}
    save_prof = ('INT', 'WIS')

    def __init__(self):
        self.pref_weapon = caster_weapons() if percent_true(75) else villager_weapons()
        self.sub_class = self.random_sub_class()
        self.name = f"Druid, Circle of {self.sub_class}"
        self.damage_type = "poison" if self.sub_class == "Spores" else damage_types("Nature")


class Fighter(Base):
    hd = 10
    fav_stat = 'STR'
    sec_stat = 'DEX'
    primary_att = 'Weapon'
    armor_ac = 15
    max_ac_dex = 2
    random_sub_class = TruffleShuffle((
        "Champion",
        "Battle Master",
        "Eldritch Knight",
        "Arcane Archer",
        "Cavalier",
        "Samurai",
        "Gladiator",
    ))
    stats = {'STR': 15, 'DEX': 13, 'CON': 14, 'INT': 8, 'WIS': 12, 'CHA': 10}
    save_prof = ('STR', 'CON')
    weapon_options = TruffleShuffle((
        lambda: f"{two_hand_weapons()}",
        lambda: f"{off_hand_shield()} and {versatile_weapons()}",
        lambda: f"{off_hand_shield()} and {main_hand_weapons()}",
        lambda: f"{off_hand_weapons()} and {main_hand_weapons()}",
        lambda: f"{off_hand_weapons()} and {off_hand_shield()}",
        lambda: f"{off_hand_weapons()} and {off_hand_weapons()}",
        lambda: f"Matched pair of {off_hand_weapons()}s",
    ))

    def __init__(self):
        self.sub_class = self.random_sub_class()
        self.name = f"Fighter, {self.sub_class}"
        if self.sub_class == "Eldritch Knight":
            self.damage_type = f"{damage_types('Physical')} and {damage_types('Arcane')}"
            self.pref_weapon = main_hand_weapons()
        elif self.sub_class == "Arcane Archer":
            self.damage_type = f"{damage_types('Physical')} and {damage_types('Arcane')}"
            self.pref_weapon = bow_weapons()
        elif self.sub_class == "Cavalier":
            self.damage_type = f"{damage_types('Physical')}"
            self.pref_weapon = pole_weapons()
        elif self.stats['DEX'] > self.stats['STR']:
            self.pref_weapon = thrown_weapons() if percent_true(50) else f"Dual Wield {off_hand_weapons()}s"
        else:
            self.damage_type = f"{damage_types('Physical')}"
            self.pref_weapon = self.weapon_options() if self.stats['STR'] > self.stats['DEX'] else thrown_weapons()


class Knight(Base):
    hd = 12
    fav_stat = 'STR'
    sec_stat = 'CON'
    primary_att = 'Polearm Weapon'
    armor_ac = 18
    max_ac_dex = 0
    random_sub_class = TruffleShuffle((
        "Flaming Tongue",
        "Frozen Wasteland",
        "Hidden Rose",
        "Storm",
        "Physical Shadow",
        "Broken Lance",
    ))
    stats = {'STR': 15, 'DEX': 14, 'CON': 13, 'INT': 12, 'WIS': 10, 'CHA': 8}
    save_prof = ('STR', 'CON')

    def __init__(self):
        self.pref_weapon = pole_weapons()
        self.sub_class = self.random_sub_class()
        self.name = f"Knight of the {self.sub_class}"
        self.damage_type = damage_types('Physical')


class Monk(Base):
    hd = 8
    fav_stat = 'DEX'
    sec_stat = 'WIS'
    primary_att = 'Unarmed Combat'
    armor_ac = 10
    max_ac_dex = 10
    random_sub_class = TruffleShuffle((
        "Open Hand",
        "Four Elements",
        "Shadow",
        "Drunken Master",
        "Kensei",
        "Sun Soul",
    ))
    stats = {'STR': 10, 'DEX': 15, 'CON': 14, 'INT': 8, 'WIS': 13, 'CHA': 12}
    save_prof = ('STR', 'DEX')
    monk_weapon = QuantumMonty(("Unarmed", "Fist Weapon", villager_weapons, off_hand_weapons)).front_linear

    def __init__(self):
        self.pref_weapon = self.monk_weapon()
        self.sub_class = self.random_sub_class()
        self.name = f"Monk, Way of the {self.sub_class}"
        if self.sub_class == "Four Elements":
            self.damage_type = f"{damage_types('Physical')} and {damage_types('Arcane')}"
        elif self.sub_class == "Shadow":
            self.damage_type = f"{damage_types('Physical')} and shadow"
        else:
            self.damage_type = f"{damage_types('Physical')}"


class Paladin(Base):
    hd = 10
    fav_stat = 'STR'
    sec_stat = 'CHA'
    primary_att = 'Melee Weapon or Divine Spell'
    armor_ac = 18
    max_ac_dex = 0
    random_sub_class = TruffleShuffle((
        "Oath of Devotion",
        "Oath of the Ancients",
        "Oath of Vengeance",
        "Oath of Conquest",
        "Oath of Redemption",
        "Oath of Glory",
        "Oath of Honor",
        "Oathbreaker",
    ))
    stats = {'STR': 15, 'DEX': 12, 'CON': 13, 'INT': 8, 'WIS': 10, 'CHA': 14}
    save_prof = ('STR', 'CHA')
    weapon_options = CumulativeWeightedChoice((
        (50, lambda: f"{pole_weapons()}"),
        (60, lambda: f"{versatile_weapons()} and {off_hand_shield()}"),
        (70, lambda: f"{two_hand_weapons()}"),
    ))

    def __init__(self):
        self.pref_weapon = self.weapon_options()
        self.sub_class = self.random_sub_class()
        self.name = f"Paladin, {self.sub_class}"
        self.damage_type = f"{damage_types('Physical')} and holy fire"


class Ranger(Base):
    hd = 8
    fav_stat = 'DEX'
    sec_stat = 'WIS'
    primary_att = 'Ranged Weapon'
    armor_ac = 12
    max_ac_dex = 10
    random_sub_class = TruffleShuffle((
        "Archer",
        "Hunter",
        "Beast Master",
        "Gloom Stalker",
        "Horizon Walker",
        "Monster Slayer",
    ))
    stats = {'STR': 8, 'DEX': 15, 'CON': 13, 'INT': 10, 'WIS': 14, 'CHA': 12}
    save_prof = ('DEX', 'WIS')
    weapon_options = QuantumMonty((
        lambda: f"{bow_weapons()}",
        lambda: f"{off_hand_weapons()} and {main_hand_weapons()}",
        lambda: f"{off_hand_weapons()} and {versatile_weapons()}",
        lambda: f"{thrown_weapons()} and {off_hand_weapons()}",
        lambda: f"Matched pair of {off_hand_weapons()}s",
        lambda: f"{two_hand_weapons()}",
    )).front_linear

    def __init__(self):
        self.sub_class = self.random_sub_class()
        self.name = f"Ranger, {self.sub_class}"
        if self.sub_class == "Archer":
            self.pref_weapon = bow_weapons()
        self.pref_weapon = self.weapon_options()
        self.damage_type = "piercing"


class Rogue(Base):
    hd = 8
    fav_stat = 'DEX'
    sec_stat = 'INT'
    primary_att = 'Light Weapon'
    armor_ac = 12
    max_ac_dex = 10
    random_sub_class = TruffleShuffle((
        "Thief",
        "Assassin",
        "Arcane Trickster",
        "Inquisitive",
        "Mastermind",
        "Scout",
        "Swashbuckler",
        "Acrobat",
    ))
    stats = {'STR': 8, 'DEX': 15, 'CON': 14, 'INT': 13, 'WIS': 12, 'CHA': 10}
    save_prof = ('DEX', 'INT')
    weapon_options = TruffleShuffle((
        lambda: f"Matched pair of {off_hand_weapons()}s",
        lambda: f"{off_hand_weapons()} and {main_hand_weapons()}",
        lambda: f"{off_hand_weapons()} and {off_hand_weapons()}",
        lambda: f"{thrown_weapons()}",
    ))

    def __init__(self):
        self.sub_class = self.random_sub_class()
        self.name = f"Rogue, {self.sub_class}"
        if self.sub_class == "Arcane Trickster":
            self.damage_type = f"{damage_types('Physical')} or {damage_types('Arcane')}"
            self.pref_weapon = off_hand_weapons() if percent_true(50) else bow_weapons()
        else:
            self.damage_type = f"{damage_types('Physical')}"
            self.pref_weapon = self.weapon_options()


class Sorcerer(Base):
    hd = 6
    name = 'Sorcerer'
    fav_stat = 'CHA'
    sec_stat = 'DEX'
    primary_att = 'Elemental Spell'
    armor_ac = 10
    max_ac_dex = 4
    random_sub_class = TruffleShuffle((
        "Fey Bloodline",
        "Dragonkin Bloodline",
        "Devilkin Bloodline",
        "Wild Magic",
        "Divine Soul",
        "Shadow Magic",
        "Storm Sorcery",
    ))
    stats = {'STR': 8, 'DEX': 14, 'CON': 13, 'INT': 12, 'WIS': 10, 'CHA': 15}
    save_prof = ('CHA', 'CON')

    def __init__(self):
        self.pref_weapon = caster_weapons() if percent_true(75) else villager_weapons()
        self.sub_class = self.random_sub_class()
        self.name = f"Sorcerer, {self.sub_class}"
        if self.sub_class == "Divine Soul":
            self.damage_type = f"{damage_types('Divine')}"
        if self.sub_class == "Shadow Magic":
            self.damage_type = "shadow"
        if self.sub_class == "Storm Sorcery":
            self.damage_type = "lightning"
        else:
            self.damage_type = f"{damage_types('Arcane')}"


class Warlock(Base):
    hd = 8
    fav_stat = 'CHA'
    sec_stat = 'DEX'
    primary_att = 'Demonic Spell'
    armor_ac = 12
    max_ac_dex = 4
    random_sub_class = TruffleShuffle((
        "Archfey",
        "Fiend",
        "Great Old One",
        "Celestial",
        "Hexblade",
    ))
    random_pact = TruffleShuffle(('Chain', 'Blade', 'Tome'))
    stats = {'STR': 8, 'DEX': 14, 'CON': 13, 'INT': 10, 'WIS': 12, 'CHA': 15}
    save_prof = ('CHA', 'WIS')

    def __init__(self):
        self.sub_class = self.random_sub_class()
        self.pact = self.random_pact()
        self.name = f"Warlock, {self.sub_class} Pact of the {self.pact}"
        if self.sub_class == "Celestial":
            self.damage_type = damage_types('Holy')
            self.pref_weapon = caster_weapons() if percent_true(75) else villager_weapons()
        else:
            self.damage_type = f"{damage_types('Arcane')}"
            self.pref_weapon = main_hand_weapons()
            if self.pact == "Blade":
                self.pref_weapon = "Longsword"
                self.damage_type += " and slashing"
            elif self.pact == "Tome":
                self.pref_weapon = "Iron Bound Tome"
                self.damage_type += " and bludgeoning"


class Wizard(Base):
    hd = 6
    name = 'Wizard'
    fav_stat = 'INT'
    sec_stat = 'DEX'
    primary_att = 'Arcane Spell'
    armor_ac = 10
    max_ac_dex = 4
    random_sub_class = TruffleShuffle((
        "Abjuration",
        "Conjuration",
        "Divination",
        "Enchantment",
        "Evocation",
        "Illusion",
        "Necromancy",
        "Transmutation",
    ))
    stats = {'STR': 8, 'DEX': 14, 'CON': 13, 'INT': 15, 'WIS': 12, 'CHA': 10}
    save_prof = ('INT', 'WIS')

    def __init__(self):
        self.pref_weapon = caster_weapons() if percent_true(75) else villager_weapons()
        self.sub_class = self.random_sub_class()
        self.name = f"Wizard, School of {self.sub_class}"
        self.damage_type = f"{damage_types('Arcane')}"


class Player:
    random_race = TruffleShuffle((
        "Human", "Mountain Dwarf", "Hill Dwarf", "Shout Halfling", "Lightfoot Halfling",
        "Forest Gnome", "Rock Gnome", "High Elf", "Wood Elf", "Half-elf", "Drow", "Half-orc",
        "Dragonborn", "Tiefling",
    ))
    racial_bonuses = {
        "Human": {"STR": 1, 'DEX': 1, 'CON': 1, 'INT': 1, 'WIS': 1, 'CHA': 1},
        "Mountain Dwarf": {"STR": 2, 'CON': 2},
        "Hill Dwarf": {"WIS": 1, 'CON': 2},
        "Shout Halfling": {"DEX": 2, 'CON': 1},
        "Lightfoot Halfling": {"DEX": 2, 'CHA': 1},
        "Forest Gnome": {"INT": 2, 'DEX': 1},
        "Rock Gnome": {"INT": 2, 'CON': 2},
        "High Elf": {"DEX": 2, 'INT': 1},
        "Wood Elf": {"DEX": 2, 'WIS': 1},
        "Half-elf": {"CHA": 2, "DEX": 1, "CON": 1},
        "Drow": {"DEX": 2, 'CHA': 1},
        "Half-orc": {"STR": 2, 'CON': 1},
        "Dragonborn": {"STR": 2, 'CHA': 1},
        "Tiefling": {"CHA": 2, 'INT': 1},
    }

    def __init__(self, cls=None, level=0, race=None, point_buy=False):
        if not cls:
            self.cls = random_character_class()() if level > 0 else Apprentice()
        else:
            self.cls = cls() if level > 0 else Apprentice()
        self.level = smart_clamp(level, 0, 20)
        self.race = self.random_race() if not race else race
        self.health = 0
        self.xp = 0
        self.ac = 10
        self.prof_bonus = 0
        self.magic_bonus = 0
        self.save_dc = 0
        self.att_bonus = 0
        self.stats = dict(self.cls.stats) if point_buy else self.set_random_stats()
        self.set_racial_bonuses()
        self.stat_mod = {}
        self.stat_mod_str = {}
        self.set_stat_mods()
        self.set_ac()
        self.set_xp()
        self.set_health()
        self.prof_bonus_str = ""
        self.set_prof()
        self.set_save_dc()
        self.set_att_bonus()
        self.saves = list(self.cls.save_prof)
        self.saving_throws = self.set_saves()
        self.background = random_background()
        self.loot = Loot()
        dam_mod = smart_clamp(self.level // 4, 1, 4)
        dam_bonus = self.stat_mod[self.cls.fav_stat] + self.magic_bonus
        self.average_damage = (dam_mod * (self.cls.damage_dice // 2)) + dam_bonus
        self.damage_formula = f"{dam_mod}d10 + {dam_bonus}"
        self.special_damage_formula = f"{smart_clamp(self.level // 2, 1, 10)}d8 + {10 + dam_bonus}"
        self.damage_type = self.find_dam_type()
        self.special_damage_type = self.cls.damage_type
        self.loot = get_loot(self.level)
        self.treasures = set()
        if self.level > 0:
            for _ in range(smart_clamp(self.level - 5, 1, 10)):
                if percent_true(1):
                    self.treasures.add(magic_table_by_cr(self.level + 10))
                elif percent_true(50):
                    self.treasures.add(magic_table_by_cr(self.level))
                elif len(self.treasures) < 1:
                    self.treasures.add(magic_table_by_cr(self.level - 10))

    def find_dam_type(self):
        dam_type = damage_types()
        for key, val in damage_type_cat.items():
            if any(ele in self.cls.pref_weapon for ele in val):
                dam_type = key
                break
        if dam_type in ("Physical", "Arcane", "Divine"):
            dam_type = damage_types(dam_type)
        return dam_type

    def set_saves(self):
        result = []
        for key, val in self.stat_mod.items():
            if key in self.cls.save_prof:
                if val + self.prof_bonus > 0:
                    result.append(f"{key} +{val + self.prof_bonus}")
                else:
                    result.append(f"{key} {val + self.prof_bonus}")
            else:
                if val > 0:
                    result.append(f"{key} +{val}")
                else:
                    result.append(f"{key} {val}")
        return result

    def set_magic(self):
        return smart_clamp(self.prof_bonus - 2 + plus_or_minus_linear(1), 0, 3) if self.level > 1 else 0

    def evaluate(self):
        return (self.level * 10) + self.att_bonus

    def __lt__(self, other):
        return self.evaluate() < other.evaluate()

    @property
    def ability_scores(self):
        return "\n".join(f"  {key} {val} ({self.stat_mod_str[key]})" for key, val in self.stats.items())

    def __repr__(self):
        n_items = len(self.treasures)
        gear = '\n  ' + f"\n  ".join(self.treasures) if n_items > 1 else list(self.treasures)[0] if n_items == 1 else 'None'
        output = (
            f"{self.cls.name}",
            f"Level: {self.level}",
            f"Total XP: {self.xp}",
            f"Race: {self.race}",
            f"Background: {self.background}",
            f"Abilities:\n" + self.ability_scores,
            f"Proficiency Bonus: {self.prof_bonus_str}",
            f"Hit Points: {self.health}",
            f"Armor Class: {self.ac}",
            f"Attack Bonus: {self.att_bonus}",
            f"Basic Damage Formula: {self.damage_formula}",
            f"Special Damage Formula: {self.special_damage_formula}",
            f"Save DC: {self.cls.fav_stat} {self.save_dc}",
            f"Save Proficiencies: {', '.join(self.saves)}",
            f"Save Modifiers: {', '.join(self.saving_throws)}",
            f"Inventory: {self.loot}",
            f"Equipment: {gear}",
            "",
        )
        return '\n'.join(output)

    def set_racial_bonuses(self):
        if self.race in self.racial_bonuses.keys():
            for stat, val in self.racial_bonuses[self.race].items():
                self.stats[stat] += val

    def set_random_stats(self):
        result = {}
        stats = ('STR', 'DEX', 'CON', 'INT', 'WIS', 'CHA')
        for ability in stats:
            if ability == self.cls.fav_stat:
                result[ability] = ability_dice(6)
            elif ability == self.cls.sec_stat:
                result[ability] = ability_dice(5)
            elif ability == 'CON':
                result[ability] = ability_dice(4)
            else:
                result[ability] = ability_dice(3)
        return result

    def set_ac(self):
        self.ac = self.cls.armor_ac + min(self.stat_mod['DEX'], self.cls.max_ac_dex) + self.magic_bonus

    def set_stat_mods(self):
        mod_by_stat = (
            -5, -5, -4, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7
        )
        for ability in self.stats:
            self.stat_mod[ability] = mod_by_stat[smart_clamp(self.stats[ability], 0, 24)]
            if self.stat_mod[ability] > 0:
                self.stat_mod_str[ability] = f"+{self.stat_mod[ability]}"
            else:
                self.stat_mod_str[ability] = f"{self.stat_mod[ability]}"

    def set_health(self):
        bonus = self.stat_mod['CON'] * self.level
        ave_hp_by_level = (self.level - 1) * ((self.cls.hd // 2) + 1)
        self.health = self.cls.hd + ave_hp_by_level + bonus

    def set_prof(self):
        prof_by_level = (0, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6)
        self.prof_bonus = prof_by_level[self.level]
        self.magic_bonus = self.set_magic()
        self.prof_bonus_str = f"{self.prof_bonus}"

    def set_save_dc(self):
        self.save_dc = 8 + self.prof_bonus + self.stat_mod[self.cls.fav_stat]

    def set_att_bonus(self):
        self.att_bonus = self.stat_mod[self.cls.fav_stat] + self.prof_bonus + self.magic_bonus

    def set_xp(self):
        xp_by_level = (
            0, 300, 900, 2700, 6500, 14000, 23000, 34000, 48000, 64000, 85000,
            100000, 120000, 140000, 165000, 195000, 225000, 265000, 305000, 355000, 400000, 800000
        )
        if self.level <= 1:
            self.xp = xp_by_level[0]
        else:
            self.xp = random_range(xp_by_level[self.level - 1], xp_by_level[self.level])
            for _ in range(self.level // 4):
                self.boost_stats()

    def boost_stats(self):
        if self.stats[self.cls.fav_stat] <= 22:
            self.stats[self.cls.fav_stat] += 2
        elif self.stats[self.cls.fav_stat] == 23:
            self.stats[self.cls.fav_stat] += 1
            self.stats[self.cls.sec_stat] += 1
        else:
            if self.stats[self.cls.sec_stat] <= 22:
                self.stats[self.cls.sec_stat] += 2
            elif self.stats[self.cls.sec_stat] == 23:
                self.stats[self.cls.sec_stat] += 1
                if self.stats['CON'] <= 23:
                    self.stats['CON'] += 1
            else:
                self.stats['CON'] += 2 if self.stats['CON'] <= 22 else 1 if self.stats['CON'] == 23 else 0
        self.set_stat_mods()


random_character_class = TruffleShuffle((
    Barbarian, Bard, Cleric,
    Druid, Fighter, Monk,
    Paladin, Ranger, Rogue,
    Sorcerer, Warlock, Wizard,
    Knight,
), flat=False)

random_tank = QuantumMonty((Knight, Fighter, Barbarian, Paladin), flat=False).front_linear
random_healer = QuantumMonty((Cleric, Druid, Paladin), flat=False).front_linear

random_dps = FlexCat({
    "Common": (Rogue, Ranger, Wizard, Sorcerer, Warlock, Monk),
    "Uncommon": (Fighter, Barbarian, Bard, Knight),
    "Rare": (Cleric, Druid, Paladin),
}, flat=False)


class Party:

    def __init__(self, level=1, strategy="default", race=None):
        self.level = level
        self.party_list = []
        self.race = race
        if strategy == "party_1":
            self.party_1()
        elif strategy == "gold":
            self.gold_party(gold_level=10)
        elif strategy == "silver":
            self.gold_party(gold_level=5)
        elif strategy == "apprentice":
            self.apprentice()
        else:
            self.default_party()

    def apprentice(self):
        self.party_list.append(Player(Apprentice, self.level, self.race))
        self.party_list.append(Player(Apprentice, self.level, self.race))
        self.party_list.append(Player(Apprentice, self.level, self.race))
        self.party_list.append(Player(Apprentice, self.level, self.race))
        self.party_list.append(Player(Apprentice, self.level, self.race))

    def default_party(self):
        self.party_list.append(Player(random_tank(), self.level, self.race))
        self.party_list.append(Player(random_healer(), self.level, self.race))
        self.party_list.append(Player(random_dps(), self.level, self.race))
        self.party_list.append(Player(random_dps(), self.level, self.race))
        self.party_list.append(Player(random_dps(), self.level, self.race))

    def party_1(self):
        self.party_list.append(Player(Paladin, self.level, self.race))
        self.party_list.append(Player(Cleric, self.level, self.race))
        self.party_list.append(Player(Wizard, self.level, self.race))
        self.party_list.append(Player(Rogue, self.level, self.race))
        self.party_list.append(Player(Ranger, self.level, self.race))

    def gold_party(self, gold_level=10):
        tank = Player(random_tank(), self.level, self.race)
        for _ in range(gold_level):
            t = Player(random_tank(), self.level, self.race)
            if t > tank:
                tank = t
        self.party_list.append(tank)
        healer = Player(random_healer(), self.level, self.race)
        for _ in range(gold_level):
            h = Player(random_healer(), self.level, self.race)
            if h > healer:
                healer = h
        self.party_list.append(healer)
        for _ in range(3):
            dps = Player(random_dps(), self.level, self.race)
            for _ in range(gold_level):
                da = Player(random_dps(), self.level, self.race)
                if da > dps:
                    dps = da
            self.party_list.append(dps)

    def __repr__(self):
        return "\n".join(str(itm) for itm in self.party_list)


if __name__ == '__main__':
    print()
    for lev in (0, 1, 5, 10, 15, 20):
        print(f"\nParty Level: {lev}")
        print(f"---------------\n")
        pc_party = Party(lev, strategy='gold')
        print(pc_party)



from .backgrounds import *
from .cr import *
from .dungeon import *
from .gear import *
from .gear_lib import *
from .monsters import *
from .monsters_lib import *
from .npc import *
from .npc_lib import *
from .player_lib import *
from .quest import *
from .quest_lib import *
from .room_lib import *
from .skills import *
from .spells import *
from .traps import *
from .treasure import *
from .utilities import *



from Fortuna import *
from DungeonGenerator.npc_lib import npc_dict


class SkillSet:

    def __init__(self, level, prof_bonus, abilities, background):
        self.skills = set()
        for ability, score in abilities.items():
            if score > 10:
                skill_bonus = prof_bonus + (score - 10) // 2
                for _ in range(-1, level // 4):
                    if skill_bonus > 0 and percent_true(score * 2):
                        self.skills.add(f"{random_value(npc_dict['skills'][ability])} {skill_bonus}")
        for skill in background.skills:
            ability = skill[-3:]
            if ability in ("STR", "DEX", "CON", "INT", "WIS", "CHA"):
                score = abilities[ability]
                self.skills.add(skill + f" {prof_bonus + (score - 10) // 2}")

    def __repr__(self):
        return ", ".join(self.skills)



from Fortuna import *

from DungeonGenerator.cr import CR
from DungeonGenerator.monsters import CampaignBoss
from DungeonGenerator.npc import random_city_npc
from DungeonGenerator.player_lib import Party
from DungeonGenerator.room_lib import MinionRoom, VillainRoom, BossRoom
from DungeonGenerator.room_lib import Room, WildArea
from DungeonGenerator.treasure import Loot


class Area:
    rooms = []

    def __getitem__(self, item):
        return self.rooms[item]

    def __str__(self):
        return "\n".join(self.rooms)


random_cond = TruffleShuffle((
    "Destroyed",
    "Abandoned",
    "Haunted",
    "Corrupted",
    "Infested",
    "Blighted",
    "Shadow",
    "Ruined",
))

random_dungeon_place = TruffleShuffle((
    "River",
    "Mountain",
    "Forest",
    "Lake",
    "Creek",
    "Swamp",
    "Valley",
))

random_dungeon_env = TruffleShuffle((
    "Tower",
    "Palace",
    "Castle",
    "Mansion",
    "Labyrinth",
    "Manor",
))

random_post_cond = TruffleShuffle((
    "Lost Souls",
    "Forgotten",
    "Mad Wizards",
    "Misty Graves",
    "Blood Queen",
    "Bone Doctor",
    "Demon Spawn",
    "Lich King",
    "Baron Gratuitous",
))


class Dungeon(Area):
    random_name = TruffleShuffle((
        lambda: f"{rand_color()} {random_dungeon_place()} {random_dungeon_env()}",
        lambda: f"The {random_cond()} {random_dungeon_env()}",
        lambda: f"{random_dungeon_env()} of the {random_post_cond()}",
        lambda: f"The {random_cond()} {random_dungeon_env()} of the {random_post_cond()}",
    ))

    def __init__(self, cr, dungeon_name=None, dungeon_levels=1,
                 rooms_per_level=10, make_traps=True):
        self.rooms = []
        self.rooms_per_level = rooms_per_level
        self.cr = CR(cr)
        self.name = dungeon_name if dungeon_name else f"{self.random_name()}"
        self.total_xp = 0
        self.dungeon_levels = dungeon_levels
        self.total_rooms = dungeon_levels * self.rooms_per_level
        room_number = 1
        first_room = MinionRoom(
            cr,
            room_name="Dungeon Entrance",
        )
        self.first_room_cr = first_room.threat.cr
        self.boss_cr = 0
        self.rooms.append(f'{first_room}')
        self.threats = [first_room]
        my_room = Room()
        for dun_level in range(dungeon_levels):
            for _ in range(self.rooms_per_level):
                room_number += 1
                if room_number == self.total_rooms:
                    my_room = BossRoom(dun_level + self.cr.value + 2, "Dungeon Heart")
                    self.boss_cr = my_room.threat.cr
                elif room_number < self.total_rooms:
                    if room_number % self.rooms_per_level == 0:
                        my_room = VillainRoom(dun_level + self.cr.value + 1)
                    elif room_number % self.rooms_per_level == 1:
                        my_room = MinionRoom(dun_level + self.cr.value)
                    elif room_number < self.total_rooms:
                        my_room = Room(dun_level + self.cr.value, make_traps=make_traps)
                else:
                    break
                self.rooms.append(f'Room {room_number}, {my_room}')
                self.threats.append(my_room)
                self.total_xp += my_room.xp
        self.total_treasure = self.get_treasures()

    def get_treasures(self):
        total = Loot()
        for room in self.threats:
            if room.threat.treasure:
                total += room.threat.treasure
        return total

    def __str__(self):
        output = (
            f"Dungeon: {self.name}\n",
            "\n".join(self.rooms),
            "\n",
        )
        return "\n".join(output)

    def __repr__(self):
        return f"Dungeon: {self.name}, {self.cr}"

    def print_room(self, room_number):
        print(self.rooms[room_number-1])


rand_color = TruffleShuffle((
    "Red",
    "Green",
    "Blue",
    "Black",
    "Shadow",
    "Ruby",
    "Emerald",
    "Jade",
    "Sapphire",
    "Diamond",
    "Onyx",
    "Golden",
    "Silver",
    "Purple",
))

rand_flavor = TruffleShuffle((
    "Rose",
    "Greenleaf",
    "Violet",
    "Lilac",
    "Flytrap",
    "Old Oak",
    "Poison",
    "Bitter",
    "Hard Candy",
    "Saffron",
    "Hot Pepper",
    "Garlic Spice",
    "Spoon",
    "Arcane",
))

random_wild = TruffleShuffle((
    TruffleShuffle(
        ("Lake", "Bay", "Swamp", "Marsh", "River", "Creek", "Island")),
    "Forest",
    "Dessert",
    "Tundra",
    "Mountain",
    "Valley",
    "Hills",
    "Cavern",
    "Canyon",
    "Swamp",
    "Wetlands",
))


class Wilderness:
    random_name = TruffleShuffle((
        lambda: f"The {rand_color()} {random_wild()}",
        lambda: f"Redwood {random_wild()}",
        lambda: f"Cavalier {random_wild()}",
        lambda: f"Frozen {random_wild()}",
        lambda: f"{random_wild()} of the Nightmare",
        lambda: f"Apple {random_wild()}",
        lambda: f"Forgotten {random_wild()}",
        lambda: f"Stormy {random_wild()}",
        lambda: f"Raven {random_wild()}",
        lambda: f"Twin Fork {random_wild()}",
        lambda: f"Twisted {random_wild()}",
        lambda: f"{random_wild()} of the Maelstrom",
        lambda: f"Titan {random_wild()}",
        lambda: f"Flaming Pitch {random_wild()}",
        lambda: f"Blighted {random_wild()}",
        lambda: f"Rainbow {random_wild()}",
        lambda: f"{rand_flavor()} {random_wild()}",
        lambda: f"Spiral {random_wild()}",
    ))

    def __init__(self, cr, name=None, num_levels=1, areas_per_level=10):
        self.cr = CR(cr)
        self.name = name if name else f"{self.random_name()}"
        self.total_xp = 0
        rooms_per_level = areas_per_level
        self.total_rooms = num_levels * rooms_per_level
        room_number = 1
        self.rooms = []
        self.threats = []
        self.total_treasure = self.get_treasures()

        for dun_level in range(num_levels):
            for _ in range(rooms_per_level):
                my_room = WildArea(dun_level + self.cr.value, self.name)
                self.rooms.append(f"Area {room_number}, {self.name}\n{my_room}")
                self.threats.append(my_room)
                if my_room.xp:
                    self.total_xp += my_room.xp
                room_number += 1

    def get_treasures(self):
        total = Loot()
        for room in self.threats:
            if room.threat.treasure:
                total += room.threat.treasure
        return total

    def __str__(self):
        output = (
            f"Wilderness: {self.name}\n",
            "\n".join(self.rooms),
            ""
        )
        return "\n".join(output)

    def __repr__(self):
        return f"Wilderness: {self.name}, {self.cr}"


random_civ = TruffleShuffle((
    "Orchards",
    "Farms",
    "Ranch",
    "Traveling Caravan",
    "Fairgrounds",
    "Outpost",
    "Hamlet",
    "Stronghold",
    "Village",
    "Township",
    "Castle",
    "Port",
    "City",
    "Citadel",
))

random_gem = TruffleShuffle((
    "Ruby",
    "Emerald",
    "Sapphire",
    "Smoky Quartz",
    "Jade",
    "Garnet",
    "Amber",
    "Pearl",
    "Moonstone",
    "Opal",
    "Topaz",
    "Sunstone",
    "Obsidian",
    "Onyx",
    "Jet",
    "Jasper",
    "Fire Agate",
    "Tiger's Eye",
    "Pyrite",
    "Moss Agate",
))

random_material = TruffleShuffle((
    "Golden",
    "Silver",
    "Iron",
    "Rusty",
    "Bronze",
    "Copper",
    "Brass",
    "Broken",
    "Glass",
    "Crystal",
    "Frozen",
    "Flaming",
))

random_thing = TruffleShuffle((
    "Sword",
    "Dagger",
    "Battleaxe",
    "Hammer",
    "Shield",
    "Saddle",
    "Lance",
    "Wand",
    "Staff",
    "Symbol",
    "Ring",
    "Amulet",
    "Suit of Armor",
    "Cannon",
    "Helm",
    "Coin",
))

random_dragon_aspect = TruffleShuffle((
    "Wing",
    "Claw",
    "Tooth",
    "Tongue",
    "Breath",
    "Tail",
    "Fist",
    "Bounty",
    "Landing",
))

random_university = TruffleShuffle((
    "Abjuration",
    "Conjuration",
    "Divination",
    "Enchantment",
    "Evocation",
    "Illusion",
    "Necromancy",
    "Transmutation",
    "War Magic",
))

random_church_type = TruffleShuffle((
    "Church",
    "Temple",
    "Shrine",
    "Cathedral",
    "Priory",
    "Chapel",
    "Altar",
    "Sanctuary",
    "Tent Revival",
))

random_light = TruffleShuffle((
    "Moon",
    "Sun",
    "Star",
    "Fire",
    "Bioluminescent",
))

random_domain = TruffleShuffle((
    "Knowledge",
    "Life",
    lambda: f"{random_light()} Light",
    "Natural Order",
    "the Tempest",
    "the Maelstrom",
    "the Jester",
    "Chivalry",
    "the Forge",
    "the Gravestone",
    "the Death Maiden",
))

random_college = TruffleShuffle((
    "Lore",
    "Valor",
    "Glamour",
    "Swords",
    "Whispers",
    "Alliteration",
))

random_monastery = TruffleShuffle((
    "the Open Hand",
    "the Four Elements",
    "Shadows",
    "the Drunken Master",
    "the Kensei Warrior",
    "the Divine Sun Soul",
    "the Crouching Tiger",
    "the Hidden Dragon",
))

random_place = TruffleShuffle((
    "Pissbottom",
    "Beggar's Hollow",
    "Rogue's Bluff",
    "Alphea",
    "Redwood",
    "Brightwater",
    "Brokenhilt",
    "Frozenleaf",
    "Riverdale",
    "Snowmelt",
    "Silverkeep",
    "Wellspring",
    "Willowdale",
    "Wolfden",
    "Shadowkeep",
))

random_school_name = TruffleShuffle((
    "Nottingham",
    "Yorkshire",
    "Bale",
    "Banford",
    "Barvard",
    "Minster & Roth",
    "Wordshire",
    "Mantisrite",
    "Alacrat",
    "Ellendale",
    "Maybereal",
    "Maybeknot",
    "Franken",
    "Nobel",
    "Daubie",
    "Kent",
    "Black Cat",
    "Canonical",
    "Spiral Tower",
))


class Settlement:
    random_name = TruffleShuffle((
        lambda: f"{random_civ()} of the {random_material()} {random_thing()}",
        lambda: f"{random_gem()} {random_civ()}",
        lambda: f"{random_place()} {random_civ()}",
        lambda: f"Monastery of {random_monastery()}",
        lambda: f"{random_school_name()} Campus: Bardic School of {random_college()}",
        lambda: f"{random_church_type()} of {random_domain()}",
        lambda: f"{random_school_name()}: University of {random_university()}",
        lambda: f"Dragon {random_dragon_aspect()} {random_civ()}",
    ))

    def __init__(self, cr, name=None, num_people=10):
        self.cr = CR(cr)
        self.num_people = num_people
        self.name = name if name else f"{self.random_name()}"
        self.rooms = []
        self.threats = []
        for room_num in range(1, self.num_people + 1):
            this_room = random_city_npc(self.cr.value)
            self.rooms.append(
                f"Area {room_num}, {self.name.split(', ')[0]}\n{this_room}")
            self.threats.append(this_room)
        self.total_treasure = self.get_treasures()

    def get_treasures(self):
        total = Loot()
        for room in self.threats:
            if room.treasure:
                total += room.treasure
        return total

    def __str__(self):
        output = (
            f"Settlement: {self.name}\n",
            "\n".join(self.rooms),
            ""
        )
        return "\n".join(output)

    def __repr__(self):
        return f"Settlement: {self.name}, {self.cr}"


class AdventureSet:

    def __init__(self, *areas):
        self.areas = areas

    def summary(self):
        return '\n'.join(f"{repr(area)}" for area in self.areas)

    def __str__(self):
        return ''.join(str(area) for area in self.areas)


def make_clichea():
    clichea = AdventureSet(
        Settlement(1, "Old Town"),
        Wilderness(1, "Gondar"),
        Dungeon(1, "Mines of Gondor"),
        Settlement(2, "Lionguard Castle"),
        Wilderness(2, "River Bridge"),
        Wilderness(3, "The Great River"),
        Wilderness(3, "Kingwood Forest"),
        Settlement(4, "Northern Shire"),
        Wilderness(4, "The Great Plain"),
        Wilderness(4, "Massive Wall"),
        Wilderness(5, "The Breach"),

        Wilderness(6, "Badlands"),
        Wilderness(7, "Land's End"),
        Wilderness(8, "World Scar"),
        Settlement(8, "Rajashi"),
        Wilderness(9, "Bandit Camp"),
        Wilderness(9, "The Reach"),
        Wilderness(10, "Dragon Tail Islands"),
        Wilderness(11, "Frozen North Wastes"),
        Dungeon(11, "Dragon's Lair"),

        Settlement(11, "Azgard"),
        Wilderness(12, "Mt. Kazzakrad"),
        Wilderness(13, "Mountains of Mist"),
        Settlement(13, "Forgehold"),
        Dungeon(14, "Vikingheim"),
        Wilderness(14, "The Maelstorm"),
        Wilderness(15, "Storm Bay"),
        Wilderness(15, "The Pass"),
        Wilderness(16, "Elvenhome"),
        Settlement(16, "Yggdraseal"),

        Settlement(17, "Lithdinlor"),
        Wilderness(17, "Star Lake"),
        Dungeon(18, "The Stones of Prophecy"),
        Wilderness(19, "Gothmor"),
        Wilderness(20, "The Dark Forest"),
        Wilderness(21, "Blackwater"),
        Wilderness(22, "Shadow Mountains"),
        Dungeon(23, "Mount Death"),
        Dungeon(24, "Dark Tower"),
        CampaignBoss(25, "The Prince of Lies"),
    )
    return clichea


def make_campaign(starting_level=1):
    adventure = AdventureSet(
        Settlement(starting_level),
        Wilderness(starting_level),
        Dungeon(starting_level+1),
        Wilderness(starting_level+2),
        Dungeon(starting_level+3),
        Wilderness(starting_level+4),
        CampaignBoss(starting_level+5),
    )
    return adventure


if __name__ == "__main__":
    print("\nPre-rolled party:\n")
    print(Party())
    campaign = make_campaign(1)
    print("\nCampaign\n")
    print(campaign)
    print("\nCampaign Summary\n")
    print(campaign.summary())
    # campaign = make_clichea()
    # print(campaign.summary(), end="\n\n")
    # print(campaign)
